from __future__ import annotations

import warnings
from typing import Any, Literal, Optional, Union

from cognite.client import data_modeling as dm
from cognite.client.data_classes import TimeSeries as CogniteTimeSeries
from pydantic import Field

from ._core import (
    DEFAULT_INSTANCE_SPACE,
    DataRecordWrite,
    DomainModel,
    DomainModelCore,
    DomainModelWrite,
    DomainModelWriteList,
    DomainModelList,
    DomainRelationWrite,
    ResourcesWrite,
    TimeSeries,
)


__all__ = [
    "Mapping",
    "MappingWrite",
    "MappingApply",
    "MappingList",
    "MappingWriteList",
    "MappingApplyList",
    "MappingFields",
    "MappingTextFields",
]


MappingTextFields = Literal["shop_path", "timeseries", "retrieve", "aggregation"]
MappingFields = Literal["shop_path", "timeseries", "transformations", "retrieve", "aggregation"]

_MAPPING_PROPERTIES_BY_FIELD = {
    "shop_path": "shopPath",
    "timeseries": "timeseries",
    "transformations": "transformations",
    "retrieve": "retrieve",
    "aggregation": "aggregation",
}


class Mapping(DomainModel):
    """This represents the reading version of mapping.

    It is used to when data is retrieved from CDF.

    Args:
        space: The space where the node is located.
        external_id: The external id of the mapping.
        data_record: The data record of the mapping node.
        shop_path: The key in shop file to map to
        timeseries: The time series to map to
        transformations: The transformations to apply to the time series
        retrieve: How to retrieve time series data
        aggregation: How to aggregate time series data
    """

    space: str = DEFAULT_INSTANCE_SPACE
    node_type: Union[dm.DirectRelationReference, None] = dm.DirectRelationReference("sp_powerops_types", "Mapping")
    shop_path: str = Field(alias="shopPath")
    timeseries: Union[TimeSeries, str, None] = None
    transformations: Optional[list[dict]] = None
    retrieve: Optional[str] = None
    aggregation: Optional[str] = None

    def as_write(self) -> MappingWrite:
        """Convert this read version of mapping to the writing version."""
        return MappingWrite(
            space=self.space,
            external_id=self.external_id,
            data_record=DataRecordWrite(existing_version=self.data_record.version),
            shop_path=self.shop_path,
            timeseries=self.timeseries,
            transformations=self.transformations,
            retrieve=self.retrieve,
            aggregation=self.aggregation,
        )

    def as_apply(self) -> MappingWrite:
        """Convert this read version of mapping to the writing version."""
        warnings.warn(
            "as_apply is deprecated and will be removed in v1.0. Use as_write instead.",
            UserWarning,
            stacklevel=2,
        )
        return self.as_write()


class MappingWrite(DomainModelWrite):
    """This represents the writing version of mapping.

    It is used to when data is sent to CDF.

    Args:
        space: The space where the node is located.
        external_id: The external id of the mapping.
        data_record: The data record of the mapping node.
        shop_path: The key in shop file to map to
        timeseries: The time series to map to
        transformations: The transformations to apply to the time series
        retrieve: How to retrieve time series data
        aggregation: How to aggregate time series data
    """

    space: str = DEFAULT_INSTANCE_SPACE
    node_type: Union[dm.DirectRelationReference, None] = dm.DirectRelationReference("sp_powerops_types", "Mapping")
    shop_path: str = Field(alias="shopPath")
    timeseries: Union[TimeSeries, str, None] = None
    transformations: Optional[list[dict]] = None
    retrieve: Optional[str] = None
    aggregation: Optional[str] = None

    def _to_instances_write(
        self,
        cache: set[tuple[str, str]],
        view_by_read_class: dict[type[DomainModelCore], dm.ViewId] | None,
        write_none: bool = False,
    ) -> ResourcesWrite:
        resources = ResourcesWrite()
        if self.as_tuple_id() in cache:
            return resources

        write_view = (view_by_read_class or {}).get(Mapping, dm.ViewId("sp_powerops_models", "Mapping", "1"))

        properties: dict[str, Any] = {}

        if self.shop_path is not None:
            properties["shopPath"] = self.shop_path

        if self.timeseries is not None or write_none:
            if isinstance(self.timeseries, str) or self.timeseries is None:
                properties["timeseries"] = self.timeseries
            else:
                properties["timeseries"] = self.timeseries.external_id

        if self.transformations is not None or write_none:
            properties["transformations"] = self.transformations

        if self.retrieve is not None or write_none:
            properties["retrieve"] = self.retrieve

        if self.aggregation is not None or write_none:
            properties["aggregation"] = self.aggregation

        if properties:
            this_node = dm.NodeApply(
                space=self.space,
                external_id=self.external_id,
                existing_version=self.data_record.existing_version,
                type=self.node_type,
                sources=[
                    dm.NodeOrEdgeData(
                        source=write_view,
                        properties=properties,
                    )
                ],
            )
            resources.nodes.append(this_node)
            cache.add(self.as_tuple_id())

        if isinstance(self.timeseries, CogniteTimeSeries):
            resources.time_series.append(self.timeseries)

        return resources


class MappingApply(MappingWrite):
    def __new__(cls, *args, **kwargs) -> MappingApply:
        warnings.warn(
            "MappingApply is deprecated and will be removed in v1.0. Use MappingWrite instead."
            "The motivation for this change is that Write is a more descriptive name for the writing version of the"
            "Mapping.",
            UserWarning,
            stacklevel=2,
        )
        return super().__new__(cls)


class MappingList(DomainModelList[Mapping]):
    """List of mappings in the read version."""

    _INSTANCE = Mapping

    def as_write(self) -> MappingWriteList:
        """Convert these read versions of mapping to the writing versions."""
        return MappingWriteList([node.as_write() for node in self.data])

    def as_apply(self) -> MappingWriteList:
        """Convert these read versions of primitive nullable to the writing versions."""
        warnings.warn(
            "as_apply is deprecated and will be removed in v1.0. Use as_write instead.",
            UserWarning,
            stacklevel=2,
        )
        return self.as_write()


class MappingWriteList(DomainModelWriteList[MappingWrite]):
    """List of mappings in the writing version."""

    _INSTANCE = MappingWrite


class MappingApplyList(MappingWriteList): ...


def _create_mapping_filter(
    view_id: dm.ViewId,
    shop_path: str | list[str] | None = None,
    shop_path_prefix: str | None = None,
    retrieve: str | list[str] | None = None,
    retrieve_prefix: str | None = None,
    aggregation: str | list[str] | None = None,
    aggregation_prefix: str | None = None,
    external_id_prefix: str | None = None,
    space: str | list[str] | None = None,
    filter: dm.Filter | None = None,
) -> dm.Filter | None:
    filters = []
    if isinstance(shop_path, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("shopPath"), value=shop_path))
    if shop_path and isinstance(shop_path, list):
        filters.append(dm.filters.In(view_id.as_property_ref("shopPath"), values=shop_path))
    if shop_path_prefix is not None:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("shopPath"), value=shop_path_prefix))
    if isinstance(retrieve, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("retrieve"), value=retrieve))
    if retrieve and isinstance(retrieve, list):
        filters.append(dm.filters.In(view_id.as_property_ref("retrieve"), values=retrieve))
    if retrieve_prefix is not None:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("retrieve"), value=retrieve_prefix))
    if isinstance(aggregation, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("aggregation"), value=aggregation))
    if aggregation and isinstance(aggregation, list):
        filters.append(dm.filters.In(view_id.as_property_ref("aggregation"), values=aggregation))
    if aggregation_prefix is not None:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("aggregation"), value=aggregation_prefix))
    if external_id_prefix is not None:
        filters.append(dm.filters.Prefix(["node", "externalId"], value=external_id_prefix))
    if isinstance(space, str):
        filters.append(dm.filters.Equals(["node", "space"], value=space))
    if space and isinstance(space, list):
        filters.append(dm.filters.In(["node", "space"], values=space))
    if filter:
        filters.append(filter)
    return dm.filters.And(*filters) if filters else None



from __future__ import annotations

import json
from typing import Optional

from pydantic import ConfigDict, Field, field_validator

from cognite.powerops.resync.config.market._core import Configuration, RelativeTime


class BenchmarkingConfig(Configuration):
    model_config = ConfigDict(populate_by_name=True)
    bid_date: RelativeTime
    shop_start: RelativeTime = Field(alias="shop_starttime")
    shop_end: RelativeTime = Field(alias="shop_endtime")
    production_plan_time_series: Optional[dict[str, list[str]]] = Field(
        default_factory=dict, alias="bid_production_plan_time_series"
    )
    market_config_external_id: str = Field(alias="bid_market_config_external_id")
    bid_process_configuration_assets: list[str] = []
    relevant_shop_objective_metrics: dict[str, str] = {
        "grand_total": "Grand Total",
        "total": "Total",
        "sum_penalties": "Sum Penalties",
        "major_penalties": "Major Penalties",
        "minor_penalties": "Minor Penalties",
        "load_value": "Load Value",
        "market_sale_buy": "Market Sale Buy",
        "rsv_end_value_relative": "RSV End Value Relative",
        "startup_costs": "Startup Costs",
        "vow_in_transit": "Vow in Transit",
        "sum_feeding_fee": "Sum Feeding Fee",
        "rsv_tactical_penalty": "RSV Tactical Penalty",
        "rsv_end_value": "RSV End Value",
        "bypass_cost": "Bypass Cost",
        "gate_discharge_cost": "Gate Discharge Cost",
        "reserve_violation_penalty": "Reserve Violation Penalty",
        "load_penalty": "Load Penalty",
    }  # Pydantic handles mutable defaults such that this is OK:
    # https://stackoverflow.com/questions/63793662/how-to-give-a-pydantic-list-field-a-default-value/63808835#63808835

    # TODO: Consider adding relationships to bid process config
    #  assets (or remove the optional part that uses those relationships in power-ops-functions)

    @field_validator("shop_start", "shop_end", "bid_date", mode="before")
    def json_loads(cls, value):
        return {"operations": json.loads(value)} if isinstance(value, str) else value



from __future__ import annotations

from dataclasses import dataclass
from typing import Any

from cognite.client.data_classes import FileMetadata
from typing_extensions import Self

from cognite.powerops.client.shop.utils import external_id_to_name

try:
    from enum import StrEnum
except ImportError:
    from strenum import StrEnum


class SHOPFileType(StrEnum):
    ASCII = "ascii"
    YAML = "yaml"


@dataclass
class SHOPFileReference:
    external_id: str
    file_type: SHOPFileType

    @property
    def name(self) -> str:
        return external_id_to_name(self.external_id)

    @classmethod
    def load(cls, data: dict[str, Any]) -> Self:
        return cls(external_id=data["external_id"], file_type=data["file_type"])

    def dump(self) -> dict[str, Any]:
        return {"external_id": self.external_id, "file_type": self.file_type}

    def as_cdf_file_metadata(self) -> FileMetadata:
        return FileMetadata(
            external_id=self.external_id,
            name=self.name,
            metadata={"shop:file_type": self.file_type},  # TODO metadata incomplete!
        )



from __future__ import annotations

from cognite.powerops.resync.models.base import Model

from .v1 import CogShop1Asset, MarketModel, ProductionModel
from .v2 import PowerAssetModelDM

V1_MODELS: list[type[Model]] = [ProductionModel, MarketModel, CogShop1Asset]
V2_MODELS: list[type[Model]] = [PowerAssetModelDM]

__all__ = ["ProductionModel", "MarketModel", "CogShop1Asset", "PowerAssetModelDM", "V1_MODELS", "V2_MODELS"]



from __future__ import annotations

import warnings
from typing import TYPE_CHECKING, Any, Literal, Optional, Union

from cognite.client import data_modeling as dm
from cognite.client.data_classes import TimeSeries as CogniteTimeSeries
from pydantic import Field

from ._core import (
    DEFAULT_INSTANCE_SPACE,
    DataRecordWrite,
    DomainModel,
    DomainModelCore,
    DomainModelWrite,
    DomainModelWriteList,
    DomainModelList,
    DomainRelationWrite,
    ResourcesWrite,
    TimeSeries,
)
from ._shop_result import SHOPResult, SHOPResultWrite

if TYPE_CHECKING:
    from ._alert import Alert, AlertWrite
    from ._case import Case, CaseWrite
    from ._shop_time_series import SHOPTimeSeries, SHOPTimeSeriesWrite


__all__ = [
    "SHOPResultPriceProd",
    "SHOPResultPriceProdWrite",
    "SHOPResultPriceProdApply",
    "SHOPResultPriceProdList",
    "SHOPResultPriceProdWriteList",
    "SHOPResultPriceProdApplyList",
    "SHOPResultPriceProdFields",
    "SHOPResultPriceProdTextFields",
]


SHOPResultPriceProdTextFields = Literal[
    "output_timeseries", "objective_sequence", "pre_run", "post_run", "shop_messages", "cplex_logs"
]
SHOPResultPriceProdFields = Literal[
    "output_timeseries", "objective_sequence", "pre_run", "post_run", "shop_messages", "cplex_logs"
]

_SHOPRESULTPRICEPROD_PROPERTIES_BY_FIELD = {
    "output_timeseries": "outputTimeseries",
    "objective_sequence": "objectiveSequence",
    "pre_run": "preRun",
    "post_run": "postRun",
    "shop_messages": "shopMessages",
    "cplex_logs": "cplexLogs",
}


class SHOPResultPriceProd(SHOPResult):
    """This represents the reading version of shop result price prod.

    It is used to when data is retrieved from CDF.

    Args:
        space: The space where the node is located.
        external_id: The external id of the shop result price prod.
        data_record: The data record of the shop result price prod node.
        case: The case that was used to produce this result
        output_timeseries: A general placeholder for all timeseries that stem from a shop run
        objective_sequence: The sequence of the objective function
        pre_run: The pre-run data for the SHOP run
        post_run: The post-run data for the SHOP run
        shop_messages: The messages from the SHOP run
        cplex_logs: The logs from CPLEX
        alerts: An array of calculation level Alerts.
        price_timeseries: The market price timeseries from the Shop run
        production_timeseries: The production timeseries wrapped as a ShopTimeSeries object containing properties related to their names and types in the resulting output shop file
    """

    node_type: Union[dm.DirectRelationReference, None] = dm.DirectRelationReference(
        "sp_powerops_types", "SHOPResultPriceProd"
    )
    price_timeseries: Union[SHOPTimeSeries, str, dm.NodeId, None] = Field(None, repr=False, alias="priceTimeseries")
    production_timeseries: Union[list[SHOPTimeSeries], list[str], None] = Field(
        default=None, repr=False, alias="productionTimeseries"
    )

    def as_write(self) -> SHOPResultPriceProdWrite:
        """Convert this read version of shop result price prod to the writing version."""
        return SHOPResultPriceProdWrite(
            space=self.space,
            external_id=self.external_id,
            data_record=DataRecordWrite(existing_version=self.data_record.version),
            case=self.case.as_write() if isinstance(self.case, DomainModel) else self.case,
            output_timeseries=self.output_timeseries,
            objective_sequence=self.objective_sequence,
            pre_run=self.pre_run,
            post_run=self.post_run,
            shop_messages=self.shop_messages,
            cplex_logs=self.cplex_logs,
            alerts=[alert.as_write() if isinstance(alert, DomainModel) else alert for alert in self.alerts or []],
            price_timeseries=(
                self.price_timeseries.as_write()
                if isinstance(self.price_timeseries, DomainModel)
                else self.price_timeseries
            ),
            production_timeseries=[
                production_timesery.as_write() if isinstance(production_timesery, DomainModel) else production_timesery
                for production_timesery in self.production_timeseries or []
            ],
        )

    def as_apply(self) -> SHOPResultPriceProdWrite:
        """Convert this read version of shop result price prod to the writing version."""
        warnings.warn(
            "as_apply is deprecated and will be removed in v1.0. Use as_write instead.",
            UserWarning,
            stacklevel=2,
        )
        return self.as_write()


class SHOPResultPriceProdWrite(SHOPResultWrite):
    """This represents the writing version of shop result price prod.

    It is used to when data is sent to CDF.

    Args:
        space: The space where the node is located.
        external_id: The external id of the shop result price prod.
        data_record: The data record of the shop result price prod node.
        case: The case that was used to produce this result
        output_timeseries: A general placeholder for all timeseries that stem from a shop run
        objective_sequence: The sequence of the objective function
        pre_run: The pre-run data for the SHOP run
        post_run: The post-run data for the SHOP run
        shop_messages: The messages from the SHOP run
        cplex_logs: The logs from CPLEX
        alerts: An array of calculation level Alerts.
        price_timeseries: The market price timeseries from the Shop run
        production_timeseries: The production timeseries wrapped as a ShopTimeSeries object containing properties related to their names and types in the resulting output shop file
    """

    node_type: Union[dm.DirectRelationReference, None] = dm.DirectRelationReference(
        "sp_powerops_types", "SHOPResultPriceProd"
    )
    price_timeseries: Union[SHOPTimeSeriesWrite, str, dm.NodeId, None] = Field(
        None, repr=False, alias="priceTimeseries"
    )
    production_timeseries: Union[list[SHOPTimeSeriesWrite], list[str], None] = Field(
        default=None, repr=False, alias="productionTimeseries"
    )

    def _to_instances_write(
        self,
        cache: set[tuple[str, str]],
        view_by_read_class: dict[type[DomainModelCore], dm.ViewId] | None,
        write_none: bool = False,
    ) -> ResourcesWrite:
        resources = ResourcesWrite()
        if self.as_tuple_id() in cache:
            return resources

        write_view = (view_by_read_class or {}).get(
            SHOPResultPriceProd, dm.ViewId("sp_powerops_models", "SHOPResultPriceProd", "1")
        )

        properties: dict[str, Any] = {}

        if self.case is not None:
            properties["case"] = {
                "space": self.space if isinstance(self.case, str) else self.case.space,
                "externalId": self.case if isinstance(self.case, str) else self.case.external_id,
            }

        if self.output_timeseries is not None or write_none:
            properties["outputTimeseries"] = [
                value if isinstance(value, str) else value.external_id for value in self.output_timeseries or []
            ] or None

        if self.objective_sequence is not None or write_none:
            properties["objectiveSequence"] = self.objective_sequence

        if self.pre_run is not None or write_none:
            properties["preRun"] = self.pre_run

        if self.post_run is not None or write_none:
            properties["postRun"] = self.post_run

        if self.shop_messages is not None or write_none:
            properties["shopMessages"] = self.shop_messages

        if self.cplex_logs is not None or write_none:
            properties["cplexLogs"] = self.cplex_logs

        if self.price_timeseries is not None:
            properties["priceTimeseries"] = {
                "space": self.space if isinstance(self.price_timeseries, str) else self.price_timeseries.space,
                "externalId": (
                    self.price_timeseries
                    if isinstance(self.price_timeseries, str)
                    else self.price_timeseries.external_id
                ),
            }

        if properties:
            this_node = dm.NodeApply(
                space=self.space,
                external_id=self.external_id,
                existing_version=self.data_record.existing_version,
                type=self.node_type,
                sources=[
                    dm.NodeOrEdgeData(
                        source=write_view,
                        properties=properties,
                    )
                ],
            )
            resources.nodes.append(this_node)
            cache.add(self.as_tuple_id())

        edge_type = dm.DirectRelationReference("sp_powerops_types", "calculationIssue")
        for alert in self.alerts or []:
            other_resources = DomainRelationWrite.from_edge_to_resources(
                cache, start_node=self, end_node=alert, edge_type=edge_type, view_by_read_class=view_by_read_class
            )
            resources.extend(other_resources)

        edge_type = dm.DirectRelationReference("sp_powerops_types", "SHOPResultPriceProd.productionTimeseries")
        for production_timesery in self.production_timeseries or []:
            other_resources = DomainRelationWrite.from_edge_to_resources(
                cache,
                start_node=self,
                end_node=production_timesery,
                edge_type=edge_type,
                view_by_read_class=view_by_read_class,
            )
            resources.extend(other_resources)

        if isinstance(self.case, DomainModelWrite):
            other_resources = self.case._to_instances_write(cache, view_by_read_class)
            resources.extend(other_resources)

        if isinstance(self.price_timeseries, DomainModelWrite):
            other_resources = self.price_timeseries._to_instances_write(cache, view_by_read_class)
            resources.extend(other_resources)

        if isinstance(self.output_timeseries, CogniteTimeSeries):
            resources.time_series.append(self.output_timeseries)

        return resources


class SHOPResultPriceProdApply(SHOPResultPriceProdWrite):
    def __new__(cls, *args, **kwargs) -> SHOPResultPriceProdApply:
        warnings.warn(
            "SHOPResultPriceProdApply is deprecated and will be removed in v1.0. Use SHOPResultPriceProdWrite instead."
            "The motivation for this change is that Write is a more descriptive name for the writing version of the"
            "SHOPResultPriceProd.",
            UserWarning,
            stacklevel=2,
        )
        return super().__new__(cls)


class SHOPResultPriceProdList(DomainModelList[SHOPResultPriceProd]):
    """List of shop result price prods in the read version."""

    _INSTANCE = SHOPResultPriceProd

    def as_write(self) -> SHOPResultPriceProdWriteList:
        """Convert these read versions of shop result price prod to the writing versions."""
        return SHOPResultPriceProdWriteList([node.as_write() for node in self.data])

    def as_apply(self) -> SHOPResultPriceProdWriteList:
        """Convert these read versions of primitive nullable to the writing versions."""
        warnings.warn(
            "as_apply is deprecated and will be removed in v1.0. Use as_write instead.",
            UserWarning,
            stacklevel=2,
        )
        return self.as_write()


class SHOPResultPriceProdWriteList(DomainModelWriteList[SHOPResultPriceProdWrite]):
    """List of shop result price prods in the writing version."""

    _INSTANCE = SHOPResultPriceProdWrite


class SHOPResultPriceProdApplyList(SHOPResultPriceProdWriteList): ...


def _create_shop_result_price_prod_filter(
    view_id: dm.ViewId,
    case: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
    price_timeseries: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
    external_id_prefix: str | None = None,
    space: str | list[str] | None = None,
    filter: dm.Filter | None = None,
) -> dm.Filter | None:
    filters = []
    if case and isinstance(case, str):
        filters.append(
            dm.filters.Equals(
                view_id.as_property_ref("case"), value={"space": DEFAULT_INSTANCE_SPACE, "externalId": case}
            )
        )
    if case and isinstance(case, tuple):
        filters.append(
            dm.filters.Equals(view_id.as_property_ref("case"), value={"space": case[0], "externalId": case[1]})
        )
    if case and isinstance(case, list) and isinstance(case[0], str):
        filters.append(
            dm.filters.In(
                view_id.as_property_ref("case"),
                values=[{"space": DEFAULT_INSTANCE_SPACE, "externalId": item} for item in case],
            )
        )
    if case and isinstance(case, list) and isinstance(case[0], tuple):
        filters.append(
            dm.filters.In(
                view_id.as_property_ref("case"), values=[{"space": item[0], "externalId": item[1]} for item in case]
            )
        )
    if price_timeseries and isinstance(price_timeseries, str):
        filters.append(
            dm.filters.Equals(
                view_id.as_property_ref("priceTimeseries"),
                value={"space": DEFAULT_INSTANCE_SPACE, "externalId": price_timeseries},
            )
        )
    if price_timeseries and isinstance(price_timeseries, tuple):
        filters.append(
            dm.filters.Equals(
                view_id.as_property_ref("priceTimeseries"),
                value={"space": price_timeseries[0], "externalId": price_timeseries[1]},
            )
        )
    if price_timeseries and isinstance(price_timeseries, list) and isinstance(price_timeseries[0], str):
        filters.append(
            dm.filters.In(
                view_id.as_property_ref("priceTimeseries"),
                values=[{"space": DEFAULT_INSTANCE_SPACE, "externalId": item} for item in price_timeseries],
            )
        )
    if price_timeseries and isinstance(price_timeseries, list) and isinstance(price_timeseries[0], tuple):
        filters.append(
            dm.filters.In(
                view_id.as_property_ref("priceTimeseries"),
                values=[{"space": item[0], "externalId": item[1]} for item in price_timeseries],
            )
        )
    if external_id_prefix is not None:
        filters.append(dm.filters.Prefix(["node", "externalId"], value=external_id_prefix))
    if isinstance(space, str):
        filters.append(dm.filters.Equals(["node", "space"], value=space))
    if space and isinstance(space, list):
        filters.append(dm.filters.In(["node", "space"], values=space))
    if filter:
        filters.append(filter)
    return dm.filters.And(*filters) if filters else None



from __future__ import annotations

from collections.abc import Sequence
from typing import overload
import warnings

from cognite.client import CogniteClient
from cognite.client import data_modeling as dm
from cognite.client.data_classes.data_modeling.instances import InstanceAggregationResultList

from cognite.powerops.client._generated.v1.data_classes._core import DEFAULT_INSTANCE_SPACE
from cognite.powerops.client._generated.v1.data_classes import (
    DomainModelCore,
    DomainModelWrite,
    ResourcesWriteResult,
    Plant,
    PlantWrite,
    PlantFields,
    PlantList,
    PlantWriteList,
    PlantTextFields,
)
from cognite.powerops.client._generated.v1.data_classes._plant import (
    _PLANT_PROPERTIES_BY_FIELD,
    _create_plant_filter,
)
from ._core import (
    DEFAULT_LIMIT_READ,
    DEFAULT_QUERY_LIMIT,
    Aggregations,
    NodeAPI,
    SequenceNotStr,
    QueryStep,
    QueryBuilder,
)
from .plant_generators import PlantGeneratorsAPI
from .plant_p_max_time_series import PlantPMaxTimeSeriesAPI
from .plant_p_min_time_series import PlantPMinTimeSeriesAPI
from .plant_water_value_time_series import PlantWaterValueTimeSeriesAPI
from .plant_feeding_fee_time_series import PlantFeedingFeeTimeSeriesAPI
from .plant_outlet_level_time_series import PlantOutletLevelTimeSeriesAPI
from .plant_inlet_level_time_series import PlantInletLevelTimeSeriesAPI
from .plant_head_direct_time_series import PlantHeadDirectTimeSeriesAPI
from .plant_query import PlantQueryAPI


class PlantAPI(NodeAPI[Plant, PlantWrite, PlantList]):
    def __init__(self, client: CogniteClient, view_by_read_class: dict[type[DomainModelCore], dm.ViewId]):
        view_id = view_by_read_class[Plant]
        super().__init__(
            client=client,
            sources=view_id,
            class_type=Plant,
            class_list=PlantList,
            class_write_list=PlantWriteList,
            view_by_read_class=view_by_read_class,
        )
        self._view_id = view_id
        self.generators_edge = PlantGeneratorsAPI(client)
        self.p_max_time_series = PlantPMaxTimeSeriesAPI(client, view_id)
        self.p_min_time_series = PlantPMinTimeSeriesAPI(client, view_id)
        self.water_value_time_series = PlantWaterValueTimeSeriesAPI(client, view_id)
        self.feeding_fee_time_series = PlantFeedingFeeTimeSeriesAPI(client, view_id)
        self.outlet_level_time_series = PlantOutletLevelTimeSeriesAPI(client, view_id)
        self.inlet_level_time_series = PlantInletLevelTimeSeriesAPI(client, view_id)
        self.head_direct_time_series = PlantHeadDirectTimeSeriesAPI(client, view_id)

    def __call__(
        self,
        name: str | list[str] | None = None,
        name_prefix: str | None = None,
        display_name: str | list[str] | None = None,
        display_name_prefix: str | None = None,
        min_ordering: int | None = None,
        max_ordering: int | None = None,
        min_head_loss_factor: float | None = None,
        max_head_loss_factor: float | None = None,
        min_outlet_level: float | None = None,
        max_outlet_level: float | None = None,
        min_p_max: float | None = None,
        max_p_max: float | None = None,
        min_p_min: float | None = None,
        max_p_min: float | None = None,
        watercourse: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        min_connection_losses: float | None = None,
        max_connection_losses: float | None = None,
        inlet_reservoir: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        external_id_prefix: str | None = None,
        space: str | list[str] | None = None,
        limit: int | None = DEFAULT_QUERY_LIMIT,
        filter: dm.Filter | None = None,
    ) -> PlantQueryAPI[PlantList]:
        """Query starting at plants.

        Args:
            name: The name to filter on.
            name_prefix: The prefix of the name to filter on.
            display_name: The display name to filter on.
            display_name_prefix: The prefix of the display name to filter on.
            min_ordering: The minimum value of the ordering to filter on.
            max_ordering: The maximum value of the ordering to filter on.
            min_head_loss_factor: The minimum value of the head loss factor to filter on.
            max_head_loss_factor: The maximum value of the head loss factor to filter on.
            min_outlet_level: The minimum value of the outlet level to filter on.
            max_outlet_level: The maximum value of the outlet level to filter on.
            min_p_max: The minimum value of the p max to filter on.
            max_p_max: The maximum value of the p max to filter on.
            min_p_min: The minimum value of the p min to filter on.
            max_p_min: The maximum value of the p min to filter on.
            watercourse: The watercourse to filter on.
            min_connection_losses: The minimum value of the connection loss to filter on.
            max_connection_losses: The maximum value of the connection loss to filter on.
            inlet_reservoir: The inlet reservoir to filter on.
            external_id_prefix: The prefix of the external ID to filter on.
            space: The space to filter on.
            limit: Maximum number of plants to return. Defaults to 25. Set to -1, float("inf") or None to return all items.
            filter: (Advanced) If the filtering available in the above is not sufficient, you can write your own filtering which will be ANDed with the filter above.

        Returns:
            A query API for plants.

        """
        has_data = dm.filters.HasData(views=[self._view_id])
        filter_ = _create_plant_filter(
            self._view_id,
            name,
            name_prefix,
            display_name,
            display_name_prefix,
            min_ordering,
            max_ordering,
            min_head_loss_factor,
            max_head_loss_factor,
            min_outlet_level,
            max_outlet_level,
            min_p_max,
            max_p_max,
            min_p_min,
            max_p_min,
            watercourse,
            min_connection_losses,
            max_connection_losses,
            inlet_reservoir,
            external_id_prefix,
            space,
            (filter and dm.filters.And(filter, has_data)) or has_data,
        )
        builder = QueryBuilder(PlantList)
        return PlantQueryAPI(self._client, builder, self._view_by_read_class, filter_, limit)

    def apply(
        self,
        plant: PlantWrite | Sequence[PlantWrite],
        replace: bool = False,
        write_none: bool = False,
    ) -> ResourcesWriteResult:
        """Add or update (upsert) plants.

        Note: This method iterates through all nodes and timeseries linked to plant and creates them including the edges
        between the nodes. For example, if any of `generators` are set, then these
        nodes as well as any nodes linked to them, and all the edges linking these nodes will be created.

        Args:
            plant: Plant or sequence of plants to upsert.
            replace (bool): How do we behave when a property value exists? Do we replace all matching and existing values with the supplied values (true)?
                Or should we merge in new values for properties together with the existing values (false)? Note: This setting applies for all nodes or edges specified in the ingestion call.
            write_none (bool): This method, will by default, skip properties that are set to None. However, if you want to set properties to None,
                you can set this parameter to True. Note this only applies to properties that are nullable.
        Returns:
            Created instance(s), i.e., nodes, edges, and time series.

        Examples:

            Create a new plant:

                >>> from cognite.powerops.client._generated.v1 import PowerOpsModelsV1Client
                >>> from cognite.powerops.client._generated.v1.data_classes import PlantWrite
                >>> client = PowerOpsModelsV1Client()
                >>> plant = PlantWrite(external_id="my_plant", ...)
                >>> result = client.plant.apply(plant)

        """
        warnings.warn(
            "The .apply method is deprecated and will be removed in v1.0. "
            "Please use the .upsert method on the client instead. This means instead of "
            "`my_client.plant.apply(my_items)` please use `my_client.upsert(my_items)`."
            "The motivation is that all apply methods are the same, and having one apply method per API "
            " class encourages users to create items in small batches, which is inefficient."
            "In addition, .upsert method is more descriptive of what the method does.",
            UserWarning,
            stacklevel=2,
        )
        return self._apply(plant, replace, write_none)

    def delete(
        self, external_id: str | SequenceNotStr[str], space: str = DEFAULT_INSTANCE_SPACE
    ) -> dm.InstancesDeleteResult:
        """Delete one or more plant.

        Args:
            external_id: External id of the plant to delete.
            space: The space where all the plant are located.

        Returns:
            The instance(s), i.e., nodes and edges which has been deleted. Empty list if nothing was deleted.

        Examples:

            Delete plant by id:

                >>> from cognite.powerops.client._generated.v1 import PowerOpsModelsV1Client
                >>> client = PowerOpsModelsV1Client()
                >>> client.plant.delete("my_plant")
        """
        warnings.warn(
            "The .delete method is deprecated and will be removed in v1.0. "
            "Please use the .delete method on the client instead. This means instead of "
            "`my_client.plant.delete(my_ids)` please use `my_client.delete(my_ids)`."
            "The motivation is that all delete methods are the same, and having one delete method per API "
            " class encourages users to delete items in small batches, which is inefficient.",
            UserWarning,
            stacklevel=2,
        )
        return self._delete(external_id, space)

    @overload
    def retrieve(self, external_id: str, space: str = DEFAULT_INSTANCE_SPACE) -> Plant | None: ...

    @overload
    def retrieve(self, external_id: SequenceNotStr[str], space: str = DEFAULT_INSTANCE_SPACE) -> PlantList: ...

    def retrieve(
        self, external_id: str | SequenceNotStr[str], space: str = DEFAULT_INSTANCE_SPACE
    ) -> Plant | PlantList | None:
        """Retrieve one or more plants by id(s).

        Args:
            external_id: External id or list of external ids of the plants.
            space: The space where all the plants are located.

        Returns:
            The requested plants.

        Examples:

            Retrieve plant by id:

                >>> from cognite.powerops.client._generated.v1 import PowerOpsModelsV1Client
                >>> client = PowerOpsModelsV1Client()
                >>> plant = client.plant.retrieve("my_plant")

        """
        return self._retrieve(
            external_id,
            space,
            retrieve_edges=True,
            edge_api_name_type_direction_view_id_penta=[
                (
                    self.generators_edge,
                    "generators",
                    dm.DirectRelationReference("sp_powerops_types", "isSubAssetOf"),
                    "outwards",
                    dm.ViewId("sp_powerops_models", "Generator", "1"),
                ),
            ],
        )

    def search(
        self,
        query: str,
        properties: PlantTextFields | Sequence[PlantTextFields] | None = None,
        name: str | list[str] | None = None,
        name_prefix: str | None = None,
        display_name: str | list[str] | None = None,
        display_name_prefix: str | None = None,
        min_ordering: int | None = None,
        max_ordering: int | None = None,
        min_head_loss_factor: float | None = None,
        max_head_loss_factor: float | None = None,
        min_outlet_level: float | None = None,
        max_outlet_level: float | None = None,
        min_p_max: float | None = None,
        max_p_max: float | None = None,
        min_p_min: float | None = None,
        max_p_min: float | None = None,
        watercourse: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        min_connection_losses: float | None = None,
        max_connection_losses: float | None = None,
        inlet_reservoir: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        external_id_prefix: str | None = None,
        space: str | list[str] | None = None,
        limit: int | None = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> PlantList:
        """Search plants

        Args:
            query: The search query,
            properties: The property to search, if nothing is passed all text fields will be searched.
            name: The name to filter on.
            name_prefix: The prefix of the name to filter on.
            display_name: The display name to filter on.
            display_name_prefix: The prefix of the display name to filter on.
            min_ordering: The minimum value of the ordering to filter on.
            max_ordering: The maximum value of the ordering to filter on.
            min_head_loss_factor: The minimum value of the head loss factor to filter on.
            max_head_loss_factor: The maximum value of the head loss factor to filter on.
            min_outlet_level: The minimum value of the outlet level to filter on.
            max_outlet_level: The maximum value of the outlet level to filter on.
            min_p_max: The minimum value of the p max to filter on.
            max_p_max: The maximum value of the p max to filter on.
            min_p_min: The minimum value of the p min to filter on.
            max_p_min: The maximum value of the p min to filter on.
            watercourse: The watercourse to filter on.
            min_connection_losses: The minimum value of the connection loss to filter on.
            max_connection_losses: The maximum value of the connection loss to filter on.
            inlet_reservoir: The inlet reservoir to filter on.
            external_id_prefix: The prefix of the external ID to filter on.
            space: The space to filter on.
            limit: Maximum number of plants to return. Defaults to 25. Set to -1, float("inf") or None to return all items.
            filter: (Advanced) If the filtering available in the above is not sufficient, you can write your own filtering which will be ANDed with the filter above.

        Returns:
            Search results plants matching the query.

        Examples:

           Search for 'my_plant' in all text properties:

                >>> from cognite.powerops.client._generated.v1 import PowerOpsModelsV1Client
                >>> client = PowerOpsModelsV1Client()
                >>> plants = client.plant.search('my_plant')

        """
        filter_ = _create_plant_filter(
            self._view_id,
            name,
            name_prefix,
            display_name,
            display_name_prefix,
            min_ordering,
            max_ordering,
            min_head_loss_factor,
            max_head_loss_factor,
            min_outlet_level,
            max_outlet_level,
            min_p_max,
            max_p_max,
            min_p_min,
            max_p_min,
            watercourse,
            min_connection_losses,
            max_connection_losses,
            inlet_reservoir,
            external_id_prefix,
            space,
            filter,
        )
        return self._search(self._view_id, query, _PLANT_PROPERTIES_BY_FIELD, properties, filter_, limit)

    @overload
    def aggregate(
        self,
        aggregations: (
            Aggregations
            | dm.aggregations.MetricAggregation
            | Sequence[Aggregations]
            | Sequence[dm.aggregations.MetricAggregation]
        ),
        property: PlantFields | Sequence[PlantFields] | None = None,
        group_by: None = None,
        query: str | None = None,
        search_properties: PlantTextFields | Sequence[PlantTextFields] | None = None,
        name: str | list[str] | None = None,
        name_prefix: str | None = None,
        display_name: str | list[str] | None = None,
        display_name_prefix: str | None = None,
        min_ordering: int | None = None,
        max_ordering: int | None = None,
        min_head_loss_factor: float | None = None,
        max_head_loss_factor: float | None = None,
        min_outlet_level: float | None = None,
        max_outlet_level: float | None = None,
        min_p_max: float | None = None,
        max_p_max: float | None = None,
        min_p_min: float | None = None,
        max_p_min: float | None = None,
        watercourse: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        min_connection_losses: float | None = None,
        max_connection_losses: float | None = None,
        inlet_reservoir: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        external_id_prefix: str | None = None,
        space: str | list[str] | None = None,
        limit: int | None = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> list[dm.aggregations.AggregatedNumberedValue]: ...

    @overload
    def aggregate(
        self,
        aggregations: (
            Aggregations
            | dm.aggregations.MetricAggregation
            | Sequence[Aggregations]
            | Sequence[dm.aggregations.MetricAggregation]
        ),
        property: PlantFields | Sequence[PlantFields] | None = None,
        group_by: PlantFields | Sequence[PlantFields] = None,
        query: str | None = None,
        search_properties: PlantTextFields | Sequence[PlantTextFields] | None = None,
        name: str | list[str] | None = None,
        name_prefix: str | None = None,
        display_name: str | list[str] | None = None,
        display_name_prefix: str | None = None,
        min_ordering: int | None = None,
        max_ordering: int | None = None,
        min_head_loss_factor: float | None = None,
        max_head_loss_factor: float | None = None,
        min_outlet_level: float | None = None,
        max_outlet_level: float | None = None,
        min_p_max: float | None = None,
        max_p_max: float | None = None,
        min_p_min: float | None = None,
        max_p_min: float | None = None,
        watercourse: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        min_connection_losses: float | None = None,
        max_connection_losses: float | None = None,
        inlet_reservoir: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        external_id_prefix: str | None = None,
        space: str | list[str] | None = None,
        limit: int | None = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> InstanceAggregationResultList: ...

    def aggregate(
        self,
        aggregate: (
            Aggregations
            | dm.aggregations.MetricAggregation
            | Sequence[Aggregations]
            | Sequence[dm.aggregations.MetricAggregation]
        ),
        property: PlantFields | Sequence[PlantFields] | None = None,
        group_by: PlantFields | Sequence[PlantFields] | None = None,
        query: str | None = None,
        search_property: PlantTextFields | Sequence[PlantTextFields] | None = None,
        name: str | list[str] | None = None,
        name_prefix: str | None = None,
        display_name: str | list[str] | None = None,
        display_name_prefix: str | None = None,
        min_ordering: int | None = None,
        max_ordering: int | None = None,
        min_head_loss_factor: float | None = None,
        max_head_loss_factor: float | None = None,
        min_outlet_level: float | None = None,
        max_outlet_level: float | None = None,
        min_p_max: float | None = None,
        max_p_max: float | None = None,
        min_p_min: float | None = None,
        max_p_min: float | None = None,
        watercourse: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        min_connection_losses: float | None = None,
        max_connection_losses: float | None = None,
        inlet_reservoir: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        external_id_prefix: str | None = None,
        space: str | list[str] | None = None,
        limit: int | None = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> list[dm.aggregations.AggregatedNumberedValue] | InstanceAggregationResultList:
        """Aggregate data across plants

        Args:
            aggregate: The aggregation to perform.
            property: The property to perform aggregation on.
            group_by: The property to group by when doing the aggregation.
            query: The query to search for in the text field.
            search_property: The text field to search in.
            name: The name to filter on.
            name_prefix: The prefix of the name to filter on.
            display_name: The display name to filter on.
            display_name_prefix: The prefix of the display name to filter on.
            min_ordering: The minimum value of the ordering to filter on.
            max_ordering: The maximum value of the ordering to filter on.
            min_head_loss_factor: The minimum value of the head loss factor to filter on.
            max_head_loss_factor: The maximum value of the head loss factor to filter on.
            min_outlet_level: The minimum value of the outlet level to filter on.
            max_outlet_level: The maximum value of the outlet level to filter on.
            min_p_max: The minimum value of the p max to filter on.
            max_p_max: The maximum value of the p max to filter on.
            min_p_min: The minimum value of the p min to filter on.
            max_p_min: The maximum value of the p min to filter on.
            watercourse: The watercourse to filter on.
            min_connection_losses: The minimum value of the connection loss to filter on.
            max_connection_losses: The maximum value of the connection loss to filter on.
            inlet_reservoir: The inlet reservoir to filter on.
            external_id_prefix: The prefix of the external ID to filter on.
            space: The space to filter on.
            limit: Maximum number of plants to return. Defaults to 25. Set to -1, float("inf") or None to return all items.
            filter: (Advanced) If the filtering available in the above is not sufficient, you can write your own filtering which will be ANDed with the filter above.

        Returns:
            Aggregation results.

        Examples:

            Count plants in space `my_space`:

                >>> from cognite.powerops.client._generated.v1 import PowerOpsModelsV1Client
                >>> client = PowerOpsModelsV1Client()
                >>> result = client.plant.aggregate("count", space="my_space")

        """

        filter_ = _create_plant_filter(
            self._view_id,
            name,
            name_prefix,
            display_name,
            display_name_prefix,
            min_ordering,
            max_ordering,
            min_head_loss_factor,
            max_head_loss_factor,
            min_outlet_level,
            max_outlet_level,
            min_p_max,
            max_p_max,
            min_p_min,
            max_p_min,
            watercourse,
            min_connection_losses,
            max_connection_losses,
            inlet_reservoir,
            external_id_prefix,
            space,
            filter,
        )
        return self._aggregate(
            self._view_id,
            aggregate,
            _PLANT_PROPERTIES_BY_FIELD,
            property,
            group_by,
            query,
            search_property,
            limit,
            filter_,
        )

    def histogram(
        self,
        property: PlantFields,
        interval: float,
        query: str | None = None,
        search_property: PlantTextFields | Sequence[PlantTextFields] | None = None,
        name: str | list[str] | None = None,
        name_prefix: str | None = None,
        display_name: str | list[str] | None = None,
        display_name_prefix: str | None = None,
        min_ordering: int | None = None,
        max_ordering: int | None = None,
        min_head_loss_factor: float | None = None,
        max_head_loss_factor: float | None = None,
        min_outlet_level: float | None = None,
        max_outlet_level: float | None = None,
        min_p_max: float | None = None,
        max_p_max: float | None = None,
        min_p_min: float | None = None,
        max_p_min: float | None = None,
        watercourse: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        min_connection_losses: float | None = None,
        max_connection_losses: float | None = None,
        inlet_reservoir: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        external_id_prefix: str | None = None,
        space: str | list[str] | None = None,
        limit: int | None = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> dm.aggregations.HistogramValue:
        """Produces histograms for plants

        Args:
            property: The property to use as the value in the histogram.
            interval: The interval to use for the histogram bins.
            query: The query to search for in the text field.
            search_property: The text field to search in.
            name: The name to filter on.
            name_prefix: The prefix of the name to filter on.
            display_name: The display name to filter on.
            display_name_prefix: The prefix of the display name to filter on.
            min_ordering: The minimum value of the ordering to filter on.
            max_ordering: The maximum value of the ordering to filter on.
            min_head_loss_factor: The minimum value of the head loss factor to filter on.
            max_head_loss_factor: The maximum value of the head loss factor to filter on.
            min_outlet_level: The minimum value of the outlet level to filter on.
            max_outlet_level: The maximum value of the outlet level to filter on.
            min_p_max: The minimum value of the p max to filter on.
            max_p_max: The maximum value of the p max to filter on.
            min_p_min: The minimum value of the p min to filter on.
            max_p_min: The maximum value of the p min to filter on.
            watercourse: The watercourse to filter on.
            min_connection_losses: The minimum value of the connection loss to filter on.
            max_connection_losses: The maximum value of the connection loss to filter on.
            inlet_reservoir: The inlet reservoir to filter on.
            external_id_prefix: The prefix of the external ID to filter on.
            space: The space to filter on.
            limit: Maximum number of plants to return. Defaults to 25. Set to -1, float("inf") or None to return all items.
            filter: (Advanced) If the filtering available in the above is not sufficient, you can write your own filtering which will be ANDed with the filter above.

        Returns:
            Bucketed histogram results.

        """
        filter_ = _create_plant_filter(
            self._view_id,
            name,
            name_prefix,
            display_name,
            display_name_prefix,
            min_ordering,
            max_ordering,
            min_head_loss_factor,
            max_head_loss_factor,
            min_outlet_level,
            max_outlet_level,
            min_p_max,
            max_p_max,
            min_p_min,
            max_p_min,
            watercourse,
            min_connection_losses,
            max_connection_losses,
            inlet_reservoir,
            external_id_prefix,
            space,
            filter,
        )
        return self._histogram(
            self._view_id,
            property,
            interval,
            _PLANT_PROPERTIES_BY_FIELD,
            query,
            search_property,
            limit,
            filter_,
        )

    def list(
        self,
        name: str | list[str] | None = None,
        name_prefix: str | None = None,
        display_name: str | list[str] | None = None,
        display_name_prefix: str | None = None,
        min_ordering: int | None = None,
        max_ordering: int | None = None,
        min_head_loss_factor: float | None = None,
        max_head_loss_factor: float | None = None,
        min_outlet_level: float | None = None,
        max_outlet_level: float | None = None,
        min_p_max: float | None = None,
        max_p_max: float | None = None,
        min_p_min: float | None = None,
        max_p_min: float | None = None,
        watercourse: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        min_connection_losses: float | None = None,
        max_connection_losses: float | None = None,
        inlet_reservoir: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        external_id_prefix: str | None = None,
        space: str | list[str] | None = None,
        limit: int | None = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
        retrieve_edges: bool = True,
    ) -> PlantList:
        """List/filter plants

        Args:
            name: The name to filter on.
            name_prefix: The prefix of the name to filter on.
            display_name: The display name to filter on.
            display_name_prefix: The prefix of the display name to filter on.
            min_ordering: The minimum value of the ordering to filter on.
            max_ordering: The maximum value of the ordering to filter on.
            min_head_loss_factor: The minimum value of the head loss factor to filter on.
            max_head_loss_factor: The maximum value of the head loss factor to filter on.
            min_outlet_level: The minimum value of the outlet level to filter on.
            max_outlet_level: The maximum value of the outlet level to filter on.
            min_p_max: The minimum value of the p max to filter on.
            max_p_max: The maximum value of the p max to filter on.
            min_p_min: The minimum value of the p min to filter on.
            max_p_min: The maximum value of the p min to filter on.
            watercourse: The watercourse to filter on.
            min_connection_losses: The minimum value of the connection loss to filter on.
            max_connection_losses: The maximum value of the connection loss to filter on.
            inlet_reservoir: The inlet reservoir to filter on.
            external_id_prefix: The prefix of the external ID to filter on.
            space: The space to filter on.
            limit: Maximum number of plants to return. Defaults to 25. Set to -1, float("inf") or None to return all items.
            filter: (Advanced) If the filtering available in the above is not sufficient, you can write your own filtering which will be ANDed with the filter above.
            retrieve_edges: Whether to retrieve `generators` external ids for the plants. Defaults to True.

        Returns:
            List of requested plants

        Examples:

            List plants and limit to 5:

                >>> from cognite.powerops.client._generated.v1 import PowerOpsModelsV1Client
                >>> client = PowerOpsModelsV1Client()
                >>> plants = client.plant.list(limit=5)

        """
        filter_ = _create_plant_filter(
            self._view_id,
            name,
            name_prefix,
            display_name,
            display_name_prefix,
            min_ordering,
            max_ordering,
            min_head_loss_factor,
            max_head_loss_factor,
            min_outlet_level,
            max_outlet_level,
            min_p_max,
            max_p_max,
            min_p_min,
            max_p_min,
            watercourse,
            min_connection_losses,
            max_connection_losses,
            inlet_reservoir,
            external_id_prefix,
            space,
            filter,
        )

        return self._list(
            limit=limit,
            filter=filter_,
            retrieve_edges=retrieve_edges,
            edge_api_name_type_direction_view_id_penta=[
                (
                    self.generators_edge,
                    "generators",
                    dm.DirectRelationReference("sp_powerops_types", "isSubAssetOf"),
                    "outwards",
                    dm.ViewId("sp_powerops_models", "Generator", "1"),
                ),
            ],
        )



from __future__ import annotations

import warnings
from typing import Any, Literal, Optional, Union

from cognite.client import data_modeling as dm
from pydantic import Field

from ._core import (
    DEFAULT_INSTANCE_SPACE,
    DataRecordWrite,
    DomainModel,
    DomainModelCore,
    DomainModelWrite,
    DomainModelWriteList,
    DomainModelList,
    DomainRelationWrite,
    ResourcesWrite,
)


__all__ = [
    "WatercourseShop",
    "WatercourseShopWrite",
    "WatercourseShopApply",
    "WatercourseShopList",
    "WatercourseShopWriteList",
    "WatercourseShopApplyList",
    "WatercourseShopFields",
    "WatercourseShopTextFields",
]


WatercourseShopTextFields = Literal["name", "display_name"]
WatercourseShopFields = Literal["name", "display_name", "ordering"]

_WATERCOURSESHOP_PROPERTIES_BY_FIELD = {
    "name": "name",
    "display_name": "displayName",
    "ordering": "ordering",
}


class WatercourseShop(DomainModel):
    """This represents the reading version of watercourse shop.

    It is used to when data is retrieved from CDF.

    Args:
        space: The space where the node is located.
        external_id: The external id of the watercourse shop.
        data_record: The data record of the watercourse shop node.
        name: Name for the Asset
        display_name: Display name for the Asset.
        ordering: The ordering of the asset
    """

    space: str = DEFAULT_INSTANCE_SPACE
    node_type: Union[dm.DirectRelationReference, None] = dm.DirectRelationReference(
        "sp_powerops_types", "WatercourseShop"
    )
    name: str
    display_name: Optional[str] = Field(None, alias="displayName")
    ordering: Optional[int] = None

    def as_write(self) -> WatercourseShopWrite:
        """Convert this read version of watercourse shop to the writing version."""
        return WatercourseShopWrite(
            space=self.space,
            external_id=self.external_id,
            data_record=DataRecordWrite(existing_version=self.data_record.version),
            name=self.name,
            display_name=self.display_name,
            ordering=self.ordering,
        )

    def as_apply(self) -> WatercourseShopWrite:
        """Convert this read version of watercourse shop to the writing version."""
        warnings.warn(
            "as_apply is deprecated and will be removed in v1.0. Use as_write instead.",
            UserWarning,
            stacklevel=2,
        )
        return self.as_write()


class WatercourseShopWrite(DomainModelWrite):
    """This represents the writing version of watercourse shop.

    It is used to when data is sent to CDF.

    Args:
        space: The space where the node is located.
        external_id: The external id of the watercourse shop.
        data_record: The data record of the watercourse shop node.
        name: Name for the Asset
        display_name: Display name for the Asset.
        ordering: The ordering of the asset
    """

    space: str = DEFAULT_INSTANCE_SPACE
    node_type: Union[dm.DirectRelationReference, None] = dm.DirectRelationReference(
        "sp_powerops_types", "WatercourseShop"
    )
    name: str
    display_name: Optional[str] = Field(None, alias="displayName")
    ordering: Optional[int] = None

    def _to_instances_write(
        self,
        cache: set[tuple[str, str]],
        view_by_read_class: dict[type[DomainModelCore], dm.ViewId] | None,
        write_none: bool = False,
    ) -> ResourcesWrite:
        resources = ResourcesWrite()
        if self.as_tuple_id() in cache:
            return resources

        write_view = (view_by_read_class or {}).get(
            WatercourseShop, dm.ViewId("sp_powerops_models", "WatercourseShop", "1")
        )

        properties: dict[str, Any] = {}

        if self.name is not None:
            properties["name"] = self.name

        if self.display_name is not None or write_none:
            properties["displayName"] = self.display_name

        if self.ordering is not None or write_none:
            properties["ordering"] = self.ordering

        if properties:
            this_node = dm.NodeApply(
                space=self.space,
                external_id=self.external_id,
                existing_version=self.data_record.existing_version,
                type=self.node_type,
                sources=[
                    dm.NodeOrEdgeData(
                        source=write_view,
                        properties=properties,
                    )
                ],
            )
            resources.nodes.append(this_node)
            cache.add(self.as_tuple_id())

        return resources


class WatercourseShopApply(WatercourseShopWrite):
    def __new__(cls, *args, **kwargs) -> WatercourseShopApply:
        warnings.warn(
            "WatercourseShopApply is deprecated and will be removed in v1.0. Use WatercourseShopWrite instead."
            "The motivation for this change is that Write is a more descriptive name for the writing version of the"
            "WatercourseShop.",
            UserWarning,
            stacklevel=2,
        )
        return super().__new__(cls)


class WatercourseShopList(DomainModelList[WatercourseShop]):
    """List of watercourse shops in the read version."""

    _INSTANCE = WatercourseShop

    def as_write(self) -> WatercourseShopWriteList:
        """Convert these read versions of watercourse shop to the writing versions."""
        return WatercourseShopWriteList([node.as_write() for node in self.data])

    def as_apply(self) -> WatercourseShopWriteList:
        """Convert these read versions of primitive nullable to the writing versions."""
        warnings.warn(
            "as_apply is deprecated and will be removed in v1.0. Use as_write instead.",
            UserWarning,
            stacklevel=2,
        )
        return self.as_write()


class WatercourseShopWriteList(DomainModelWriteList[WatercourseShopWrite]):
    """List of watercourse shops in the writing version."""

    _INSTANCE = WatercourseShopWrite


class WatercourseShopApplyList(WatercourseShopWriteList): ...


def _create_watercourse_shop_filter(
    view_id: dm.ViewId,
    name: str | list[str] | None = None,
    name_prefix: str | None = None,
    display_name: str | list[str] | None = None,
    display_name_prefix: str | None = None,
    min_ordering: int | None = None,
    max_ordering: int | None = None,
    external_id_prefix: str | None = None,
    space: str | list[str] | None = None,
    filter: dm.Filter | None = None,
) -> dm.Filter | None:
    filters = []
    if isinstance(name, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("name"), value=name))
    if name and isinstance(name, list):
        filters.append(dm.filters.In(view_id.as_property_ref("name"), values=name))
    if name_prefix is not None:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("name"), value=name_prefix))
    if isinstance(display_name, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("displayName"), value=display_name))
    if display_name and isinstance(display_name, list):
        filters.append(dm.filters.In(view_id.as_property_ref("displayName"), values=display_name))
    if display_name_prefix is not None:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("displayName"), value=display_name_prefix))
    if min_ordering is not None or max_ordering is not None:
        filters.append(dm.filters.Range(view_id.as_property_ref("ordering"), gte=min_ordering, lte=max_ordering))
    if external_id_prefix is not None:
        filters.append(dm.filters.Prefix(["node", "externalId"], value=external_id_prefix))
    if isinstance(space, str):
        filters.append(dm.filters.Equals(["node", "space"], value=space))
    if space and isinstance(space, list):
        filters.append(dm.filters.In(["node", "space"], values=space))
    if filter:
        filters.append(filter)
    return dm.filters.And(*filters) if filters else None



from __future__ import annotations

from itertools import zip_longest
from typing import Any, Literal

from cognite.client.data_classes import AssetList, LabelDefinitionList
from cognite.client.data_classes._base import CogniteResource, CogniteResourceList

from cognite.powerops.client._generated.assets.data_classes import DomainModelWrite as DomainModelApplyAssets
from cognite.powerops.client._generated.cogshop1.data_classes._core import DomainModelApply as DomainModelApplyCogShop1
from cognite.powerops.resync.models.base import AssetModel, CDFFile, CDFSequence, Model, ResourceType
from cognite.powerops.utils.serialization import remove_read_only_fields

from .data_classes import Change, FieldDifference, ModelDifference


def model_difference(
    current_model: Model, new_model: Model, static_resources: dict[str, AssetList | LabelDefinitionList] | None = None
) -> ModelDifference:
    if type(current_model) != type(new_model):
        raise ValueError(f"Cannot compare model of type {type(current_model)} with {type(new_model)}")
    # The dump and load calls are to remove all read only fields
    current_reloaded = current_model.load_from_cdf_resources(current_model.dump_as_cdf_resource(), link="external_id")
    new_reloaded = new_model.load_from_cdf_resources(new_model.dump_as_cdf_resource(), link="external_id")
    static_resources = {
        k: type(v)._load([remove_read_only_fields(resource.dump(camel_case=True)) for resource in v])
        for k, v in (static_resources or {}).items()
    }

    diffs = []
    for field_name in current_model.model_fields:
        current_value = getattr(current_reloaded, field_name)
        new_value = getattr(new_reloaded, field_name)
        if type(current_value) != type(new_value):
            raise ValueError(f"Cannot compare field {field_name} of type {type(current_value)} with {type(new_value)}")
        diff = _find_diffs(current_value, new_value, "Domain", field_name)
        diffs.append(diff)

    for function in current_model.cdf_resources:
        current_items = function(current_reloaded)
        new_items = function(new_reloaded)

        diff = _find_diffs(current_items, new_items, "CDF", function.__name__)

        diffs.append(diff)

    if isinstance(new_model, AssetModel) and static_resources:
        for field_name, static_resource in static_resources.items():
            new_resources = getattr(new_reloaded, field_name)()
            diff = _find_diffs(static_resource, new_resources, "CDF", field_name)
            if field_name == "labels":
                # Labels are not removed, only added
                diff.unchanged.extend(diff.removed)
                diff.removed = []
            diffs.append(diff)

    return ModelDifference(model_name=current_model.model_name, changes={d.field_name: d for d in diffs})


def remove_only(
    model: Model, static_resources: dict[str, LabelDefinitionList | AssetList] | None = None
) -> ModelDifference:
    # The dump and load calls are to remove all read only fields
    current_reloaded = model.load_from_cdf_resources(model.dump_as_cdf_resource())
    diffs = []
    for field_name in current_reloaded.model_fields:
        diff = _as_removals(getattr(current_reloaded, field_name), "Domain", field_name)
        diffs.append(diff)

    for function in current_reloaded.cdf_resources:
        diff = _as_removals(function(current_reloaded), "CDF", function.__name__)
        diffs.append(diff)

    for field_name, resources in (static_resources or {}).items():
        diff = _as_removals(resources, "CDF", field_name)
        diffs.append(diff)

    return ModelDifference(model_name=model.model_name, changes={d.field_name: d for d in diffs})


def _find_diffs(
    current_value: Any, new_value: Any, group: Literal["CDF", "Domain"], field_name: str
) -> FieldDifference:
    if not current_value and not new_value:
        return FieldDifference(group=group, field_name=field_name)

    current_value_by_id = _to_value_by_id(current_value)
    new_value_by_id = _to_value_by_id(new_value)

    added_ids = set(new_value_by_id.keys()) - set(current_value_by_id.keys())
    added = [new_value_by_id[external_id] for external_id in sorted(added_ids)]

    removed_ids = set(current_value_by_id.keys()) - set(new_value_by_id.keys())
    removed = [current_value_by_id[external_id] for external_id in sorted(removed_ids)]

    existing_ids = set(current_value_by_id.keys()) & set(new_value_by_id.keys())
    changed = []
    unchanged = []
    for existing in sorted(existing_ids):
        current = current_value_by_id[existing]
        new = new_value_by_id[existing]
        if _is_equals(current, new):
            unchanged.append(current)
        else:
            changed.append(Change(last=current, new=new))

    return FieldDifference(
        group=group, field_name=field_name, added=added, removed=removed, changed=changed, unchanged=unchanged
    )


def _is_equals(current: Any, new: Any) -> bool:
    if current == new:
        return True
    sentinel = (object(), object())
    for (k1, v1), (k2, v2) in zip_longest(vars(current).items(), vars(new).items(), fillvalue=sentinel):
        if k1 != k2:
            return False
        if v1 != v2 and (v1 and v2):
            return False
    return True


def _as_removals(value: Any, group: Literal["CDF", "Domain"], field_name: str) -> FieldDifference:
    if not value:
        return FieldDifference(group=group, field_name=field_name)

    value_by_id = _to_value_by_id(value)
    removed = [value_by_id[external_id] for external_id in sorted(value_by_id.keys())]
    return FieldDifference(group=group, field_name=field_name, removed=removed)


def _to_value_by_id(value: Any) -> dict[str, Any]:
    if (
        isinstance(value, (list, CogniteResourceList))
        and value
        and isinstance(value[0], (ResourceType, CogniteResource))
    ):
        return {item.external_id: item for item in value}
    elif isinstance(value, list) and value and isinstance(value[0], (CDFSequence, CDFFile)):
        return {item.external_id: item.cdf_resource for item in value}
    elif isinstance(value, dict) and value and isinstance(next(iter(value.values())), (DomainModelApplyCogShop1)):
        return {item.external_id: item for item in value.values()}
    elif isinstance(value, (dict, list, CogniteResourceList)) and not value:
        return {}
    elif isinstance(value, list) and value and isinstance(value[0], (DomainModelApplyCogShop1, DomainModelApplyAssets)):
        return {item.external_id: item for item in value}
    raise NotImplementedError(f"{type(value)} is not supported")



from __future__ import annotations

import warnings
from typing import TYPE_CHECKING, Any, Literal, Optional, Union

from cognite.client import data_modeling as dm
from pydantic import Field

from ._core import (
    DEFAULT_INSTANCE_SPACE,
    DataRecordWrite,
    DomainModel,
    DomainModelCore,
    DomainModelWrite,
    DomainModelWriteList,
    DomainModelList,
    DomainRelationWrite,
    ResourcesWrite,
)

if TYPE_CHECKING:
    from ._alert import Alert, AlertWrite
    from ._bid_document_day_ahead import BidDocumentDayAhead, BidDocumentDayAheadWrite
    from ._total_bid_matrix_calculation_input import TotalBidMatrixCalculationInput, TotalBidMatrixCalculationInputWrite


__all__ = [
    "TotalBidMatrixCalculationOutput",
    "TotalBidMatrixCalculationOutputWrite",
    "TotalBidMatrixCalculationOutputApply",
    "TotalBidMatrixCalculationOutputList",
    "TotalBidMatrixCalculationOutputWriteList",
    "TotalBidMatrixCalculationOutputApplyList",
    "TotalBidMatrixCalculationOutputFields",
    "TotalBidMatrixCalculationOutputTextFields",
]


TotalBidMatrixCalculationOutputTextFields = Literal["process_id", "function_name", "function_call_id"]
TotalBidMatrixCalculationOutputFields = Literal["process_id", "process_step", "function_name", "function_call_id"]

_TOTALBIDMATRIXCALCULATIONOUTPUT_PROPERTIES_BY_FIELD = {
    "process_id": "processId",
    "process_step": "processStep",
    "function_name": "functionName",
    "function_call_id": "functionCallId",
}


class TotalBidMatrixCalculationOutput(DomainModel):
    """This represents the reading version of total bid matrix calculation output.

    It is used to when data is retrieved from CDF.

    Args:
        space: The space where the node is located.
        external_id: The external id of the total bid matrix calculation output.
        data_record: The data record of the total bid matrix calculation output node.
        process_id: The process associated with the function execution
        process_step: This is the step in the process.
        function_name: The name of the function
        function_call_id: The function call id
        alerts: An array of calculation level Alerts.
        bid_document: The bid document field.
        input_: The previous step in the process.
    """

    space: str = DEFAULT_INSTANCE_SPACE
    node_type: Union[dm.DirectRelationReference, None] = dm.DirectRelationReference(
        "sp_powerops_types", "TotalBidMatrixCalculationOutput"
    )
    process_id: str = Field(alias="processId")
    process_step: int = Field(alias="processStep")
    function_name: str = Field(alias="functionName")
    function_call_id: str = Field(alias="functionCallId")
    alerts: Union[list[Alert], list[str], None] = Field(default=None, repr=False)
    bid_document: Union[BidDocumentDayAhead, str, dm.NodeId, None] = Field(None, repr=False, alias="bidDocument")
    input_: Union[TotalBidMatrixCalculationInput, str, dm.NodeId, None] = Field(None, repr=False, alias="input")

    def as_write(self) -> TotalBidMatrixCalculationOutputWrite:
        """Convert this read version of total bid matrix calculation output to the writing version."""
        return TotalBidMatrixCalculationOutputWrite(
            space=self.space,
            external_id=self.external_id,
            data_record=DataRecordWrite(existing_version=self.data_record.version),
            process_id=self.process_id,
            process_step=self.process_step,
            function_name=self.function_name,
            function_call_id=self.function_call_id,
            alerts=[alert.as_write() if isinstance(alert, DomainModel) else alert for alert in self.alerts or []],
            bid_document=(
                self.bid_document.as_write() if isinstance(self.bid_document, DomainModel) else self.bid_document
            ),
            input_=self.input_.as_write() if isinstance(self.input_, DomainModel) else self.input_,
        )

    def as_apply(self) -> TotalBidMatrixCalculationOutputWrite:
        """Convert this read version of total bid matrix calculation output to the writing version."""
        warnings.warn(
            "as_apply is deprecated and will be removed in v1.0. Use as_write instead.",
            UserWarning,
            stacklevel=2,
        )
        return self.as_write()


class TotalBidMatrixCalculationOutputWrite(DomainModelWrite):
    """This represents the writing version of total bid matrix calculation output.

    It is used to when data is sent to CDF.

    Args:
        space: The space where the node is located.
        external_id: The external id of the total bid matrix calculation output.
        data_record: The data record of the total bid matrix calculation output node.
        process_id: The process associated with the function execution
        process_step: This is the step in the process.
        function_name: The name of the function
        function_call_id: The function call id
        alerts: An array of calculation level Alerts.
        bid_document: The bid document field.
        input_: The previous step in the process.
    """

    space: str = DEFAULT_INSTANCE_SPACE
    node_type: Union[dm.DirectRelationReference, None] = dm.DirectRelationReference(
        "sp_powerops_types", "TotalBidMatrixCalculationOutput"
    )
    process_id: str = Field(alias="processId")
    process_step: int = Field(alias="processStep")
    function_name: str = Field(alias="functionName")
    function_call_id: str = Field(alias="functionCallId")
    alerts: Union[list[AlertWrite], list[str], None] = Field(default=None, repr=False)
    bid_document: Union[BidDocumentDayAheadWrite, str, dm.NodeId, None] = Field(None, repr=False, alias="bidDocument")
    input_: Union[TotalBidMatrixCalculationInputWrite, str, dm.NodeId, None] = Field(None, repr=False, alias="input")

    def _to_instances_write(
        self,
        cache: set[tuple[str, str]],
        view_by_read_class: dict[type[DomainModelCore], dm.ViewId] | None,
        write_none: bool = False,
    ) -> ResourcesWrite:
        resources = ResourcesWrite()
        if self.as_tuple_id() in cache:
            return resources

        write_view = (view_by_read_class or {}).get(
            TotalBidMatrixCalculationOutput, dm.ViewId("sp_powerops_models", "TotalBidMatrixCalculationOutput", "1")
        )

        properties: dict[str, Any] = {}

        if self.process_id is not None:
            properties["processId"] = self.process_id

        if self.process_step is not None:
            properties["processStep"] = self.process_step

        if self.function_name is not None:
            properties["functionName"] = self.function_name

        if self.function_call_id is not None:
            properties["functionCallId"] = self.function_call_id

        if self.bid_document is not None:
            properties["bidDocument"] = {
                "space": self.space if isinstance(self.bid_document, str) else self.bid_document.space,
                "externalId": (
                    self.bid_document if isinstance(self.bid_document, str) else self.bid_document.external_id
                ),
            }

        if self.input_ is not None:
            properties["input"] = {
                "space": self.space if isinstance(self.input_, str) else self.input_.space,
                "externalId": self.input_ if isinstance(self.input_, str) else self.input_.external_id,
            }

        if properties:
            this_node = dm.NodeApply(
                space=self.space,
                external_id=self.external_id,
                existing_version=self.data_record.existing_version,
                type=self.node_type,
                sources=[
                    dm.NodeOrEdgeData(
                        source=write_view,
                        properties=properties,
                    )
                ],
            )
            resources.nodes.append(this_node)
            cache.add(self.as_tuple_id())

        edge_type = dm.DirectRelationReference("sp_powerops_types", "calculationIssue")
        for alert in self.alerts or []:
            other_resources = DomainRelationWrite.from_edge_to_resources(
                cache, start_node=self, end_node=alert, edge_type=edge_type, view_by_read_class=view_by_read_class
            )
            resources.extend(other_resources)

        if isinstance(self.bid_document, DomainModelWrite):
            other_resources = self.bid_document._to_instances_write(cache, view_by_read_class)
            resources.extend(other_resources)

        if isinstance(self.input_, DomainModelWrite):
            other_resources = self.input_._to_instances_write(cache, view_by_read_class)
            resources.extend(other_resources)

        return resources


class TotalBidMatrixCalculationOutputApply(TotalBidMatrixCalculationOutputWrite):
    def __new__(cls, *args, **kwargs) -> TotalBidMatrixCalculationOutputApply:
        warnings.warn(
            "TotalBidMatrixCalculationOutputApply is deprecated and will be removed in v1.0. Use TotalBidMatrixCalculationOutputWrite instead."
            "The motivation for this change is that Write is a more descriptive name for the writing version of the"
            "TotalBidMatrixCalculationOutput.",
            UserWarning,
            stacklevel=2,
        )
        return super().__new__(cls)


class TotalBidMatrixCalculationOutputList(DomainModelList[TotalBidMatrixCalculationOutput]):
    """List of total bid matrix calculation outputs in the read version."""

    _INSTANCE = TotalBidMatrixCalculationOutput

    def as_write(self) -> TotalBidMatrixCalculationOutputWriteList:
        """Convert these read versions of total bid matrix calculation output to the writing versions."""
        return TotalBidMatrixCalculationOutputWriteList([node.as_write() for node in self.data])

    def as_apply(self) -> TotalBidMatrixCalculationOutputWriteList:
        """Convert these read versions of primitive nullable to the writing versions."""
        warnings.warn(
            "as_apply is deprecated and will be removed in v1.0. Use as_write instead.",
            UserWarning,
            stacklevel=2,
        )
        return self.as_write()


class TotalBidMatrixCalculationOutputWriteList(DomainModelWriteList[TotalBidMatrixCalculationOutputWrite]):
    """List of total bid matrix calculation outputs in the writing version."""

    _INSTANCE = TotalBidMatrixCalculationOutputWrite


class TotalBidMatrixCalculationOutputApplyList(TotalBidMatrixCalculationOutputWriteList): ...


def _create_total_bid_matrix_calculation_output_filter(
    view_id: dm.ViewId,
    process_id: str | list[str] | None = None,
    process_id_prefix: str | None = None,
    min_process_step: int | None = None,
    max_process_step: int | None = None,
    function_name: str | list[str] | None = None,
    function_name_prefix: str | None = None,
    function_call_id: str | list[str] | None = None,
    function_call_id_prefix: str | None = None,
    bid_document: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
    input_: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
    external_id_prefix: str | None = None,
    space: str | list[str] | None = None,
    filter: dm.Filter | None = None,
) -> dm.Filter | None:
    filters = []
    if isinstance(process_id, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("processId"), value=process_id))
    if process_id and isinstance(process_id, list):
        filters.append(dm.filters.In(view_id.as_property_ref("processId"), values=process_id))
    if process_id_prefix is not None:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("processId"), value=process_id_prefix))
    if min_process_step is not None or max_process_step is not None:
        filters.append(
            dm.filters.Range(view_id.as_property_ref("processStep"), gte=min_process_step, lte=max_process_step)
        )
    if isinstance(function_name, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("functionName"), value=function_name))
    if function_name and isinstance(function_name, list):
        filters.append(dm.filters.In(view_id.as_property_ref("functionName"), values=function_name))
    if function_name_prefix is not None:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("functionName"), value=function_name_prefix))
    if isinstance(function_call_id, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("functionCallId"), value=function_call_id))
    if function_call_id and isinstance(function_call_id, list):
        filters.append(dm.filters.In(view_id.as_property_ref("functionCallId"), values=function_call_id))
    if function_call_id_prefix is not None:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("functionCallId"), value=function_call_id_prefix))
    if bid_document and isinstance(bid_document, str):
        filters.append(
            dm.filters.Equals(
                view_id.as_property_ref("bidDocument"),
                value={"space": DEFAULT_INSTANCE_SPACE, "externalId": bid_document},
            )
        )
    if bid_document and isinstance(bid_document, tuple):
        filters.append(
            dm.filters.Equals(
                view_id.as_property_ref("bidDocument"), value={"space": bid_document[0], "externalId": bid_document[1]}
            )
        )
    if bid_document and isinstance(bid_document, list) and isinstance(bid_document[0], str):
        filters.append(
            dm.filters.In(
                view_id.as_property_ref("bidDocument"),
                values=[{"space": DEFAULT_INSTANCE_SPACE, "externalId": item} for item in bid_document],
            )
        )
    if bid_document and isinstance(bid_document, list) and isinstance(bid_document[0], tuple):
        filters.append(
            dm.filters.In(
                view_id.as_property_ref("bidDocument"),
                values=[{"space": item[0], "externalId": item[1]} for item in bid_document],
            )
        )
    if input_ and isinstance(input_, str):
        filters.append(
            dm.filters.Equals(
                view_id.as_property_ref("input"), value={"space": DEFAULT_INSTANCE_SPACE, "externalId": input_}
            )
        )
    if input_ and isinstance(input_, tuple):
        filters.append(
            dm.filters.Equals(view_id.as_property_ref("input"), value={"space": input_[0], "externalId": input_[1]})
        )
    if input_ and isinstance(input_, list) and isinstance(input_[0], str):
        filters.append(
            dm.filters.In(
                view_id.as_property_ref("input"),
                values=[{"space": DEFAULT_INSTANCE_SPACE, "externalId": item} for item in input_],
            )
        )
    if input_ and isinstance(input_, list) and isinstance(input_[0], tuple):
        filters.append(
            dm.filters.In(
                view_id.as_property_ref("input"), values=[{"space": item[0], "externalId": item[1]} for item in input_]
            )
        )
    if external_id_prefix is not None:
        filters.append(dm.filters.Prefix(["node", "externalId"], value=external_id_prefix))
    if isinstance(space, str):
        filters.append(dm.filters.Equals(["node", "space"], value=space))
    if space and isinstance(space, list):
        filters.append(dm.filters.In(["node", "space"], values=space))
    if filter:
        filters.append(filter)
    return dm.filters.And(*filters) if filters else None



from pydantic import BaseModel

from cognite.powerops.client._logger import LoggingLevelT


class Settings(BaseModel):
    data_set_external_id: str
    overwrite_data: bool
    organization_subdomain: str
    tenant_id: str
    shop_version: str
    cdf_project: str
    debug_level: LoggingLevelT = "INFO"

    @property
    def shop_service_url(self) -> str:
        return (
            "https://shop-production.az-inso-powerops.cognite.ai/submit-run"
            if self.cdf_project.endswith("-prod")
            else "https://shop-staging.az-inso-powerops.cognite.ai/submit-run"
        )



from __future__ import annotations

from abc import ABC
from typing import TypeVar, Union

from cognite.client.data_classes import Asset, FileMetadata, Sequence, TimeSeries
from cognite.client.data_classes.data_modeling import ContainerId, DataModelId, ViewId
from pydantic import BaseModel
from typing_extensions import TypeAlias

from .cdf_resources import CDFFile, CDFSequence, SpaceId
from .helpers import isinstance_list

_T_Type = TypeVar("_T_Type")


class ResourceType(BaseModel, ABC):
    def sequences(self) -> list[CDFSequence]:
        return self._fields_of_type(CDFSequence)

    def files(self) -> list[CDFFile]:
        return self._fields_of_type(CDFFile)

    def time_series(self) -> list[TimeSeries]:
        return self._fields_of_type(TimeSeries)

    def _fields_of_type(self, type_: type[_T_Type]) -> list[_T_Type]:
        output: list[_T_Type] = []
        for field_name in self.model_fields:
            value = getattr(self, field_name)
            if not value:
                continue
            elif isinstance_list(value, type_):
                output.extend(value)
            elif isinstance(value, type_):
                output.append(value)
        return output

    @property
    def external_id(self) -> str:
        raise NotImplementedError()

    def standardize(self) -> None:
        """
        This ensures that the model is in a standardized form.

        This is useful when doing comparisons between models, as for example, the ordering of the assets in some
        lists does not matter.
        """
        ...


Resource: TypeAlias = Union[
    Asset, TimeSeries, Sequence, FileMetadata, ResourceType, ContainerId, ViewId, DataModelId, SpaceId
]



from __future__ import annotations

import datetime
import warnings
from typing import TYPE_CHECKING, Any, Literal, Optional, Union

from cognite.client import data_modeling as dm
from pydantic import Field

from ._core import (
    DEFAULT_INSTANCE_SPACE,
    DataRecordWrite,
    DomainModel,
    DomainModelCore,
    DomainModelWrite,
    DomainModelWriteList,
    DomainModelList,
    DomainRelationWrite,
    ResourcesWrite,
)

if TYPE_CHECKING:
    from ._scenario import Scenario, ScenarioWrite


__all__ = [
    "Case",
    "CaseWrite",
    "CaseApply",
    "CaseList",
    "CaseWriteList",
    "CaseApplyList",
    "CaseFields",
    "CaseTextFields",
]


CaseTextFields = Literal["case_file", "reservoir_mapping", "cut_order_files", "extra_files"]
CaseFields = Literal[
    "case_file",
    "reservoir_mapping",
    "cut_order_files",
    "extra_files",
    "cog_shop_files_config",
    "start_time",
    "end_time",
]

_CASE_PROPERTIES_BY_FIELD = {
    "case_file": "caseFile",
    "reservoir_mapping": "reservoirMapping",
    "cut_order_files": "cutOrderFiles",
    "extra_files": "extraFiles",
    "cog_shop_files_config": "cogShopFilesConfig",
    "start_time": "startTime",
    "end_time": "endTime",
}


class Case(DomainModel):
    """This represents the reading version of case.

    It is used to when data is retrieved from CDF.

    Args:
        space: The space where the node is located.
        external_id: The external id of the case.
        data_record: The data record of the case node.
        scenario: The Shop scenario that was used to produce this result
        case_file: The case file used
        reservoir_mapping: The cut file reservoir mapping
        cut_order_files: Cut order files (Module series in PRODRISK)
        extra_files: The extra file field.
        cog_shop_files_config: Configuration for in what order to load the various files into pyshop
        start_time: The start time of the case
        end_time: The end time of the case
    """

    space: str = DEFAULT_INSTANCE_SPACE
    node_type: Union[dm.DirectRelationReference, None] = dm.DirectRelationReference("sp_powerops_types", "Case")
    scenario: Union[Scenario, str, dm.NodeId, None] = Field(None, repr=False)
    case_file: Union[str, None] = Field(None, alias="caseFile")
    reservoir_mapping: Optional[list[str]] = Field(None, alias="reservoirMapping")
    cut_order_files: Optional[list[str]] = Field(None, alias="cutOrderFiles")
    extra_files: Optional[list[str]] = Field(None, alias="extraFiles")
    cog_shop_files_config: Optional[list[dict]] = Field(None, alias="cogShopFilesConfig")
    start_time: Optional[datetime.date] = Field(None, alias="startTime")
    end_time: Optional[datetime.date] = Field(None, alias="endTime")

    def as_write(self) -> CaseWrite:
        """Convert this read version of case to the writing version."""
        return CaseWrite(
            space=self.space,
            external_id=self.external_id,
            data_record=DataRecordWrite(existing_version=self.data_record.version),
            scenario=self.scenario.as_write() if isinstance(self.scenario, DomainModel) else self.scenario,
            case_file=self.case_file,
            reservoir_mapping=self.reservoir_mapping,
            cut_order_files=self.cut_order_files,
            extra_files=self.extra_files,
            cog_shop_files_config=self.cog_shop_files_config,
            start_time=self.start_time,
            end_time=self.end_time,
        )

    def as_apply(self) -> CaseWrite:
        """Convert this read version of case to the writing version."""
        warnings.warn(
            "as_apply is deprecated and will be removed in v1.0. Use as_write instead.",
            UserWarning,
            stacklevel=2,
        )
        return self.as_write()


class CaseWrite(DomainModelWrite):
    """This represents the writing version of case.

    It is used to when data is sent to CDF.

    Args:
        space: The space where the node is located.
        external_id: The external id of the case.
        data_record: The data record of the case node.
        scenario: The Shop scenario that was used to produce this result
        case_file: The case file used
        reservoir_mapping: The cut file reservoir mapping
        cut_order_files: Cut order files (Module series in PRODRISK)
        extra_files: The extra file field.
        cog_shop_files_config: Configuration for in what order to load the various files into pyshop
        start_time: The start time of the case
        end_time: The end time of the case
    """

    space: str = DEFAULT_INSTANCE_SPACE
    node_type: Union[dm.DirectRelationReference, None] = dm.DirectRelationReference("sp_powerops_types", "Case")
    scenario: Union[ScenarioWrite, str, dm.NodeId, None] = Field(None, repr=False)
    case_file: Union[str, None] = Field(None, alias="caseFile")
    reservoir_mapping: Optional[list[str]] = Field(None, alias="reservoirMapping")
    cut_order_files: Optional[list[str]] = Field(None, alias="cutOrderFiles")
    extra_files: Optional[list[str]] = Field(None, alias="extraFiles")
    cog_shop_files_config: Optional[list[dict]] = Field(None, alias="cogShopFilesConfig")
    start_time: Optional[datetime.date] = Field(None, alias="startTime")
    end_time: Optional[datetime.date] = Field(None, alias="endTime")

    def _to_instances_write(
        self,
        cache: set[tuple[str, str]],
        view_by_read_class: dict[type[DomainModelCore], dm.ViewId] | None,
        write_none: bool = False,
    ) -> ResourcesWrite:
        resources = ResourcesWrite()
        if self.as_tuple_id() in cache:
            return resources

        write_view = (view_by_read_class or {}).get(Case, dm.ViewId("sp_powerops_models", "Case", "1"))

        properties: dict[str, Any] = {}

        if self.scenario is not None:
            properties["scenario"] = {
                "space": self.space if isinstance(self.scenario, str) else self.scenario.space,
                "externalId": self.scenario if isinstance(self.scenario, str) else self.scenario.external_id,
            }

        if self.case_file is not None or write_none:
            properties["caseFile"] = self.case_file

        if self.reservoir_mapping is not None or write_none:
            properties["reservoirMapping"] = self.reservoir_mapping

        if self.cut_order_files is not None or write_none:
            properties["cutOrderFiles"] = self.cut_order_files

        if self.extra_files is not None or write_none:
            properties["extraFiles"] = self.extra_files

        if self.cog_shop_files_config is not None or write_none:
            properties["cogShopFilesConfig"] = self.cog_shop_files_config

        if self.start_time is not None or write_none:
            properties["startTime"] = self.start_time.isoformat() if self.start_time else None

        if self.end_time is not None or write_none:
            properties["endTime"] = self.end_time.isoformat() if self.end_time else None

        if properties:
            this_node = dm.NodeApply(
                space=self.space,
                external_id=self.external_id,
                existing_version=self.data_record.existing_version,
                type=self.node_type,
                sources=[
                    dm.NodeOrEdgeData(
                        source=write_view,
                        properties=properties,
                    )
                ],
            )
            resources.nodes.append(this_node)
            cache.add(self.as_tuple_id())

        if isinstance(self.scenario, DomainModelWrite):
            other_resources = self.scenario._to_instances_write(cache, view_by_read_class)
            resources.extend(other_resources)

        return resources


class CaseApply(CaseWrite):
    def __new__(cls, *args, **kwargs) -> CaseApply:
        warnings.warn(
            "CaseApply is deprecated and will be removed in v1.0. Use CaseWrite instead."
            "The motivation for this change is that Write is a more descriptive name for the writing version of the"
            "Case.",
            UserWarning,
            stacklevel=2,
        )
        return super().__new__(cls)


class CaseList(DomainModelList[Case]):
    """List of cases in the read version."""

    _INSTANCE = Case

    def as_write(self) -> CaseWriteList:
        """Convert these read versions of case to the writing versions."""
        return CaseWriteList([node.as_write() for node in self.data])

    def as_apply(self) -> CaseWriteList:
        """Convert these read versions of primitive nullable to the writing versions."""
        warnings.warn(
            "as_apply is deprecated and will be removed in v1.0. Use as_write instead.",
            UserWarning,
            stacklevel=2,
        )
        return self.as_write()


class CaseWriteList(DomainModelWriteList[CaseWrite]):
    """List of cases in the writing version."""

    _INSTANCE = CaseWrite


class CaseApplyList(CaseWriteList): ...


def _create_case_filter(
    view_id: dm.ViewId,
    scenario: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
    min_start_time: datetime.date | None = None,
    max_start_time: datetime.date | None = None,
    min_end_time: datetime.date | None = None,
    max_end_time: datetime.date | None = None,
    external_id_prefix: str | None = None,
    space: str | list[str] | None = None,
    filter: dm.Filter | None = None,
) -> dm.Filter | None:
    filters = []
    if scenario and isinstance(scenario, str):
        filters.append(
            dm.filters.Equals(
                view_id.as_property_ref("scenario"), value={"space": DEFAULT_INSTANCE_SPACE, "externalId": scenario}
            )
        )
    if scenario and isinstance(scenario, tuple):
        filters.append(
            dm.filters.Equals(
                view_id.as_property_ref("scenario"), value={"space": scenario[0], "externalId": scenario[1]}
            )
        )
    if scenario and isinstance(scenario, list) and isinstance(scenario[0], str):
        filters.append(
            dm.filters.In(
                view_id.as_property_ref("scenario"),
                values=[{"space": DEFAULT_INSTANCE_SPACE, "externalId": item} for item in scenario],
            )
        )
    if scenario and isinstance(scenario, list) and isinstance(scenario[0], tuple):
        filters.append(
            dm.filters.In(
                view_id.as_property_ref("scenario"),
                values=[{"space": item[0], "externalId": item[1]} for item in scenario],
            )
        )
    if min_start_time is not None or max_start_time is not None:
        filters.append(
            dm.filters.Range(
                view_id.as_property_ref("startTime"),
                gte=min_start_time.isoformat() if min_start_time else None,
                lte=max_start_time.isoformat() if max_start_time else None,
            )
        )
    if min_end_time is not None or max_end_time is not None:
        filters.append(
            dm.filters.Range(
                view_id.as_property_ref("endTime"),
                gte=min_end_time.isoformat() if min_end_time else None,
                lte=max_end_time.isoformat() if max_end_time else None,
            )
        )
    if external_id_prefix is not None:
        filters.append(dm.filters.Prefix(["node", "externalId"], value=external_id_prefix))
    if isinstance(space, str):
        filters.append(dm.filters.Equals(["node", "space"], value=space))
    if space and isinstance(space, list):
        filters.append(dm.filters.In(["node", "space"], values=space))
    if filter:
        filters.append(filter)
    return dm.filters.And(*filters) if filters else None



from ._core import (
    DataRecord,
    DataRecordWrite,
    DomainModel,
    DomainModelCore,
    DomainModelWrite,
    DomainModelList,
    DomainRelationWrite,
    ResourcesWrite,
    ResourcesWriteResult,
)
from ._alert import (
    Alert,
    AlertApply,
    AlertApplyList,
    AlertFields,
    AlertList,
    AlertTextFields,
    AlertWrite,
    AlertWriteList,
)
from ._basic_bid_matrix import (
    BasicBidMatrix,
    BasicBidMatrixApply,
    BasicBidMatrixApplyList,
    BasicBidMatrixFields,
    BasicBidMatrixList,
    BasicBidMatrixTextFields,
    BasicBidMatrixWrite,
    BasicBidMatrixWriteList,
)
from ._bid_calculation_task import (
    BidCalculationTask,
    BidCalculationTaskApply,
    BidCalculationTaskApplyList,
    BidCalculationTaskFields,
    BidCalculationTaskList,
    BidCalculationTaskWrite,
    BidCalculationTaskWriteList,
)
from ._bid_configuration import (
    BidConfiguration,
    BidConfigurationApply,
    BidConfigurationApplyList,
    BidConfigurationList,
    BidConfigurationWrite,
    BidConfigurationWriteList,
)
from ._bid_configuration_shop import (
    BidConfigurationShop,
    BidConfigurationShopApply,
    BidConfigurationShopApplyList,
    BidConfigurationShopFields,
    BidConfigurationShopList,
    BidConfigurationShopTextFields,
    BidConfigurationShopWrite,
    BidConfigurationShopWriteList,
)
from ._bid_configuration_water import (
    BidConfigurationWater,
    BidConfigurationWaterApply,
    BidConfigurationWaterApplyList,
    BidConfigurationWaterList,
    BidConfigurationWaterWrite,
    BidConfigurationWaterWriteList,
)
from ._bid_document_afrr import (
    BidDocumentAFRR,
    BidDocumentAFRRApply,
    BidDocumentAFRRApplyList,
    BidDocumentAFRRFields,
    BidDocumentAFRRList,
    BidDocumentAFRRTextFields,
    BidDocumentAFRRWrite,
    BidDocumentAFRRWriteList,
)
from ._bid_document_day_ahead import (
    BidDocumentDayAhead,
    BidDocumentDayAheadApply,
    BidDocumentDayAheadApplyList,
    BidDocumentDayAheadFields,
    BidDocumentDayAheadList,
    BidDocumentDayAheadTextFields,
    BidDocumentDayAheadWrite,
    BidDocumentDayAheadWriteList,
)
from ._bid_matrix import (
    BidMatrix,
    BidMatrixApply,
    BidMatrixApplyList,
    BidMatrixFields,
    BidMatrixList,
    BidMatrixTextFields,
    BidMatrixWrite,
    BidMatrixWriteList,
)
from ._bid_matrix_raw import (
    BidMatrixRaw,
    BidMatrixRawApply,
    BidMatrixRawApplyList,
    BidMatrixRawFields,
    BidMatrixRawList,
    BidMatrixRawTextFields,
    BidMatrixRawWrite,
    BidMatrixRawWriteList,
)
from ._bid_method import (
    BidMethod,
    BidMethodApply,
    BidMethodApplyList,
    BidMethodFields,
    BidMethodList,
    BidMethodTextFields,
    BidMethodWrite,
    BidMethodWriteList,
)
from ._bid_method_afrr import (
    BidMethodAFRR,
    BidMethodAFRRApply,
    BidMethodAFRRApplyList,
    BidMethodAFRRFields,
    BidMethodAFRRList,
    BidMethodAFRRTextFields,
    BidMethodAFRRWrite,
    BidMethodAFRRWriteList,
)
from ._bid_method_custom import (
    BidMethodCustom,
    BidMethodCustomApply,
    BidMethodCustomApplyList,
    BidMethodCustomFields,
    BidMethodCustomList,
    BidMethodCustomTextFields,
    BidMethodCustomWrite,
    BidMethodCustomWriteList,
)
from ._bid_method_day_ahead import (
    BidMethodDayAhead,
    BidMethodDayAheadApply,
    BidMethodDayAheadApplyList,
    BidMethodDayAheadFields,
    BidMethodDayAheadList,
    BidMethodDayAheadTextFields,
    BidMethodDayAheadWrite,
    BidMethodDayAheadWriteList,
)
from ._bid_method_shop_multi_scenario import (
    BidMethodSHOPMultiScenario,
    BidMethodSHOPMultiScenarioApply,
    BidMethodSHOPMultiScenarioApplyList,
    BidMethodSHOPMultiScenarioFields,
    BidMethodSHOPMultiScenarioList,
    BidMethodSHOPMultiScenarioTextFields,
    BidMethodSHOPMultiScenarioWrite,
    BidMethodSHOPMultiScenarioWriteList,
)
from ._bid_method_water_value import (
    BidMethodWaterValue,
    BidMethodWaterValueApply,
    BidMethodWaterValueApplyList,
    BidMethodWaterValueFields,
    BidMethodWaterValueList,
    BidMethodWaterValueTextFields,
    BidMethodWaterValueWrite,
    BidMethodWaterValueWriteList,
)
from ._bid_row import (
    BidRow,
    BidRowApply,
    BidRowApplyList,
    BidRowFields,
    BidRowList,
    BidRowTextFields,
    BidRowWrite,
    BidRowWriteList,
)
from ._case import Case, CaseApply, CaseApplyList, CaseFields, CaseList, CaseWrite, CaseWriteList
from ._commands import (
    Commands,
    CommandsApply,
    CommandsApplyList,
    CommandsFields,
    CommandsList,
    CommandsTextFields,
    CommandsWrite,
    CommandsWriteList,
)
from ._custom_bid_matrix import (
    CustomBidMatrix,
    CustomBidMatrixApply,
    CustomBidMatrixApplyList,
    CustomBidMatrixFields,
    CustomBidMatrixList,
    CustomBidMatrixTextFields,
    CustomBidMatrixWrite,
    CustomBidMatrixWriteList,
)
from ._generator import (
    Generator,
    GeneratorApply,
    GeneratorApplyList,
    GeneratorFields,
    GeneratorList,
    GeneratorTextFields,
    GeneratorWrite,
    GeneratorWriteList,
)
from ._generator_efficiency_curve import (
    GeneratorEfficiencyCurve,
    GeneratorEfficiencyCurveApply,
    GeneratorEfficiencyCurveApplyList,
    GeneratorEfficiencyCurveFields,
    GeneratorEfficiencyCurveList,
    GeneratorEfficiencyCurveWrite,
    GeneratorEfficiencyCurveWriteList,
)
from ._mapping import (
    Mapping,
    MappingApply,
    MappingApplyList,
    MappingFields,
    MappingList,
    MappingTextFields,
    MappingWrite,
    MappingWriteList,
)
from ._market_configuration import (
    MarketConfiguration,
    MarketConfigurationApply,
    MarketConfigurationApplyList,
    MarketConfigurationFields,
    MarketConfigurationList,
    MarketConfigurationTextFields,
    MarketConfigurationWrite,
    MarketConfigurationWriteList,
)
from ._model_template import (
    ModelTemplate,
    ModelTemplateApply,
    ModelTemplateApplyList,
    ModelTemplateFields,
    ModelTemplateList,
    ModelTemplateTextFields,
    ModelTemplateWrite,
    ModelTemplateWriteList,
)
from ._multi_scenario_matrix import (
    MultiScenarioMatrix,
    MultiScenarioMatrixApply,
    MultiScenarioMatrixApplyList,
    MultiScenarioMatrixFields,
    MultiScenarioMatrixList,
    MultiScenarioMatrixTextFields,
    MultiScenarioMatrixWrite,
    MultiScenarioMatrixWriteList,
)
from ._multi_scenario_matrix_raw import (
    MultiScenarioMatrixRaw,
    MultiScenarioMatrixRawApply,
    MultiScenarioMatrixRawApplyList,
    MultiScenarioMatrixRawFields,
    MultiScenarioMatrixRawList,
    MultiScenarioMatrixRawTextFields,
    MultiScenarioMatrixRawWrite,
    MultiScenarioMatrixRawWriteList,
)
from ._partial_post_processing_input import (
    PartialPostProcessingInput,
    PartialPostProcessingInputApply,
    PartialPostProcessingInputApplyList,
    PartialPostProcessingInputFields,
    PartialPostProcessingInputList,
    PartialPostProcessingInputTextFields,
    PartialPostProcessingInputWrite,
    PartialPostProcessingInputWriteList,
)
from ._partial_post_processing_output import (
    PartialPostProcessingOutput,
    PartialPostProcessingOutputApply,
    PartialPostProcessingOutputApplyList,
    PartialPostProcessingOutputFields,
    PartialPostProcessingOutputList,
    PartialPostProcessingOutputTextFields,
    PartialPostProcessingOutputWrite,
    PartialPostProcessingOutputWriteList,
)
from ._plant import (
    Plant,
    PlantApply,
    PlantApplyList,
    PlantFields,
    PlantList,
    PlantTextFields,
    PlantWrite,
    PlantWriteList,
)
from ._plant_shop import (
    PlantShop,
    PlantShopApply,
    PlantShopApplyList,
    PlantShopFields,
    PlantShopList,
    PlantShopTextFields,
    PlantShopWrite,
    PlantShopWriteList,
)
from ._preprocessor_input import (
    PreprocessorInput,
    PreprocessorInputApply,
    PreprocessorInputApplyList,
    PreprocessorInputFields,
    PreprocessorInputList,
    PreprocessorInputTextFields,
    PreprocessorInputWrite,
    PreprocessorInputWriteList,
)
from ._preprocessor_output import (
    PreprocessorOutput,
    PreprocessorOutputApply,
    PreprocessorOutputApplyList,
    PreprocessorOutputFields,
    PreprocessorOutputList,
    PreprocessorOutputTextFields,
    PreprocessorOutputWrite,
    PreprocessorOutputWriteList,
)
from ._price_area import (
    PriceArea,
    PriceAreaApply,
    PriceAreaApplyList,
    PriceAreaFields,
    PriceAreaList,
    PriceAreaTextFields,
    PriceAreaWrite,
    PriceAreaWriteList,
)
from ._price_area_afrr import (
    PriceAreaAFRR,
    PriceAreaAFRRApply,
    PriceAreaAFRRApplyList,
    PriceAreaAFRRFields,
    PriceAreaAFRRList,
    PriceAreaAFRRTextFields,
    PriceAreaAFRRWrite,
    PriceAreaAFRRWriteList,
)
from ._price_area_asset import (
    PriceAreaAsset,
    PriceAreaAssetApply,
    PriceAreaAssetApplyList,
    PriceAreaAssetFields,
    PriceAreaAssetList,
    PriceAreaAssetTextFields,
    PriceAreaAssetWrite,
    PriceAreaAssetWriteList,
)
from ._price_prod_case import (
    PriceProdCase,
    PriceProdCaseApply,
    PriceProdCaseApplyList,
    PriceProdCaseFields,
    PriceProdCaseList,
    PriceProdCaseWrite,
    PriceProdCaseWriteList,
)
from ._reservoir import (
    Reservoir,
    ReservoirApply,
    ReservoirApplyList,
    ReservoirFields,
    ReservoirList,
    ReservoirTextFields,
    ReservoirWrite,
    ReservoirWriteList,
)
from ._shop_result import (
    SHOPResult,
    SHOPResultApply,
    SHOPResultApplyList,
    SHOPResultFields,
    SHOPResultList,
    SHOPResultWrite,
    SHOPResultWriteList,
)
from ._shop_result_price_prod import (
    SHOPResultPriceProd,
    SHOPResultPriceProdApply,
    SHOPResultPriceProdApplyList,
    SHOPResultPriceProdFields,
    SHOPResultPriceProdList,
    SHOPResultPriceProdWrite,
    SHOPResultPriceProdWriteList,
)
from ._shop_time_series import (
    SHOPTimeSeries,
    SHOPTimeSeriesApply,
    SHOPTimeSeriesApplyList,
    SHOPTimeSeriesFields,
    SHOPTimeSeriesList,
    SHOPTimeSeriesTextFields,
    SHOPTimeSeriesWrite,
    SHOPTimeSeriesWriteList,
)
from ._shop_trigger_input import (
    SHOPTriggerInput,
    SHOPTriggerInputApply,
    SHOPTriggerInputApplyList,
    SHOPTriggerInputFields,
    SHOPTriggerInputList,
    SHOPTriggerInputTextFields,
    SHOPTriggerInputWrite,
    SHOPTriggerInputWriteList,
)
from ._shop_trigger_output import (
    SHOPTriggerOutput,
    SHOPTriggerOutputApply,
    SHOPTriggerOutputApplyList,
    SHOPTriggerOutputFields,
    SHOPTriggerOutputList,
    SHOPTriggerOutputTextFields,
    SHOPTriggerOutputWrite,
    SHOPTriggerOutputWriteList,
)
from ._scenario import (
    Scenario,
    ScenarioApply,
    ScenarioApplyList,
    ScenarioFields,
    ScenarioList,
    ScenarioTextFields,
    ScenarioWrite,
    ScenarioWriteList,
)
from ._shop_partial_bid_calculation_input import (
    ShopPartialBidCalculationInput,
    ShopPartialBidCalculationInputApply,
    ShopPartialBidCalculationInputApplyList,
    ShopPartialBidCalculationInputFields,
    ShopPartialBidCalculationInputList,
    ShopPartialBidCalculationInputTextFields,
    ShopPartialBidCalculationInputWrite,
    ShopPartialBidCalculationInputWriteList,
)
from ._shop_partial_bid_calculation_output import (
    ShopPartialBidCalculationOutput,
    ShopPartialBidCalculationOutputApply,
    ShopPartialBidCalculationOutputApplyList,
    ShopPartialBidCalculationOutputFields,
    ShopPartialBidCalculationOutputList,
    ShopPartialBidCalculationOutputTextFields,
    ShopPartialBidCalculationOutputWrite,
    ShopPartialBidCalculationOutputWriteList,
)
from ._task_dispatcher_shop_input import (
    TaskDispatcherShopInput,
    TaskDispatcherShopInputApply,
    TaskDispatcherShopInputApplyList,
    TaskDispatcherShopInputFields,
    TaskDispatcherShopInputList,
    TaskDispatcherShopInputTextFields,
    TaskDispatcherShopInputWrite,
    TaskDispatcherShopInputWriteList,
)
from ._task_dispatcher_shop_output import (
    TaskDispatcherShopOutput,
    TaskDispatcherShopOutputApply,
    TaskDispatcherShopOutputApplyList,
    TaskDispatcherShopOutputFields,
    TaskDispatcherShopOutputList,
    TaskDispatcherShopOutputTextFields,
    TaskDispatcherShopOutputWrite,
    TaskDispatcherShopOutputWriteList,
)
from ._task_dispatcher_water_input import (
    TaskDispatcherWaterInput,
    TaskDispatcherWaterInputApply,
    TaskDispatcherWaterInputApplyList,
    TaskDispatcherWaterInputFields,
    TaskDispatcherWaterInputList,
    TaskDispatcherWaterInputTextFields,
    TaskDispatcherWaterInputWrite,
    TaskDispatcherWaterInputWriteList,
)
from ._task_dispatcher_water_output import (
    TaskDispatcherWaterOutput,
    TaskDispatcherWaterOutputApply,
    TaskDispatcherWaterOutputApplyList,
    TaskDispatcherWaterOutputFields,
    TaskDispatcherWaterOutputList,
    TaskDispatcherWaterOutputTextFields,
    TaskDispatcherWaterOutputWrite,
    TaskDispatcherWaterOutputWriteList,
)
from ._total_bid_matrix_calculation_input import (
    TotalBidMatrixCalculationInput,
    TotalBidMatrixCalculationInputApply,
    TotalBidMatrixCalculationInputApplyList,
    TotalBidMatrixCalculationInputFields,
    TotalBidMatrixCalculationInputList,
    TotalBidMatrixCalculationInputTextFields,
    TotalBidMatrixCalculationInputWrite,
    TotalBidMatrixCalculationInputWriteList,
)
from ._total_bid_matrix_calculation_output import (
    TotalBidMatrixCalculationOutput,
    TotalBidMatrixCalculationOutputApply,
    TotalBidMatrixCalculationOutputApplyList,
    TotalBidMatrixCalculationOutputFields,
    TotalBidMatrixCalculationOutputList,
    TotalBidMatrixCalculationOutputTextFields,
    TotalBidMatrixCalculationOutputWrite,
    TotalBidMatrixCalculationOutputWriteList,
)
from ._turbine_efficiency_curve import (
    TurbineEfficiencyCurve,
    TurbineEfficiencyCurveApply,
    TurbineEfficiencyCurveApplyList,
    TurbineEfficiencyCurveFields,
    TurbineEfficiencyCurveList,
    TurbineEfficiencyCurveWrite,
    TurbineEfficiencyCurveWriteList,
)
from ._water_partial_bid_calculation_input import (
    WaterPartialBidCalculationInput,
    WaterPartialBidCalculationInputApply,
    WaterPartialBidCalculationInputApplyList,
    WaterPartialBidCalculationInputFields,
    WaterPartialBidCalculationInputList,
    WaterPartialBidCalculationInputTextFields,
    WaterPartialBidCalculationInputWrite,
    WaterPartialBidCalculationInputWriteList,
)
from ._water_partial_bid_calculation_output import (
    WaterPartialBidCalculationOutput,
    WaterPartialBidCalculationOutputApply,
    WaterPartialBidCalculationOutputApplyList,
    WaterPartialBidCalculationOutputFields,
    WaterPartialBidCalculationOutputList,
    WaterPartialBidCalculationOutputTextFields,
    WaterPartialBidCalculationOutputWrite,
    WaterPartialBidCalculationOutputWriteList,
)
from ._watercourse import (
    Watercourse,
    WatercourseApply,
    WatercourseApplyList,
    WatercourseFields,
    WatercourseList,
    WatercourseTextFields,
    WatercourseWrite,
    WatercourseWriteList,
)
from ._watercourse_shop import (
    WatercourseShop,
    WatercourseShopApply,
    WatercourseShopApplyList,
    WatercourseShopFields,
    WatercourseShopList,
    WatercourseShopTextFields,
    WatercourseShopWrite,
    WatercourseShopWriteList,
)

BasicBidMatrix.model_rebuild()
BasicBidMatrixWrite.model_rebuild()
BasicBidMatrixApply.model_rebuild()
BidCalculationTask.model_rebuild()
BidCalculationTaskWrite.model_rebuild()
BidCalculationTaskApply.model_rebuild()
BidConfiguration.model_rebuild()
BidConfigurationWrite.model_rebuild()
BidConfigurationApply.model_rebuild()
BidConfigurationShop.model_rebuild()
BidConfigurationShopWrite.model_rebuild()
BidConfigurationShopApply.model_rebuild()
BidConfigurationWater.model_rebuild()
BidConfigurationWaterWrite.model_rebuild()
BidConfigurationWaterApply.model_rebuild()
BidDocumentAFRR.model_rebuild()
BidDocumentAFRRWrite.model_rebuild()
BidDocumentAFRRApply.model_rebuild()
BidDocumentDayAhead.model_rebuild()
BidDocumentDayAheadWrite.model_rebuild()
BidDocumentDayAheadApply.model_rebuild()
BidMatrix.model_rebuild()
BidMatrixWrite.model_rebuild()
BidMatrixApply.model_rebuild()
BidMatrixRaw.model_rebuild()
BidMatrixRawWrite.model_rebuild()
BidMatrixRawApply.model_rebuild()
BidMethodSHOPMultiScenario.model_rebuild()
BidMethodSHOPMultiScenarioWrite.model_rebuild()
BidMethodSHOPMultiScenarioApply.model_rebuild()
BidRow.model_rebuild()
BidRowWrite.model_rebuild()
BidRowApply.model_rebuild()
Case.model_rebuild()
CaseWrite.model_rebuild()
CaseApply.model_rebuild()
CustomBidMatrix.model_rebuild()
CustomBidMatrixWrite.model_rebuild()
CustomBidMatrixApply.model_rebuild()
Generator.model_rebuild()
GeneratorWrite.model_rebuild()
GeneratorApply.model_rebuild()
ModelTemplate.model_rebuild()
ModelTemplateWrite.model_rebuild()
ModelTemplateApply.model_rebuild()
MultiScenarioMatrix.model_rebuild()
MultiScenarioMatrixWrite.model_rebuild()
MultiScenarioMatrixApply.model_rebuild()
MultiScenarioMatrixRaw.model_rebuild()
MultiScenarioMatrixRawWrite.model_rebuild()
MultiScenarioMatrixRawApply.model_rebuild()
PartialPostProcessingInput.model_rebuild()
PartialPostProcessingInputWrite.model_rebuild()
PartialPostProcessingInputApply.model_rebuild()
PartialPostProcessingOutput.model_rebuild()
PartialPostProcessingOutputWrite.model_rebuild()
PartialPostProcessingOutputApply.model_rebuild()
Plant.model_rebuild()
PlantWrite.model_rebuild()
PlantApply.model_rebuild()
PreprocessorInput.model_rebuild()
PreprocessorInputWrite.model_rebuild()
PreprocessorInputApply.model_rebuild()
PreprocessorOutput.model_rebuild()
PreprocessorOutputWrite.model_rebuild()
PreprocessorOutputApply.model_rebuild()
PriceAreaAsset.model_rebuild()
PriceAreaAssetWrite.model_rebuild()
PriceAreaAssetApply.model_rebuild()
PriceProdCase.model_rebuild()
PriceProdCaseWrite.model_rebuild()
PriceProdCaseApply.model_rebuild()
SHOPResult.model_rebuild()
SHOPResultWrite.model_rebuild()
SHOPResultApply.model_rebuild()
SHOPResultPriceProd.model_rebuild()
SHOPResultPriceProdWrite.model_rebuild()
SHOPResultPriceProdApply.model_rebuild()
SHOPTriggerInput.model_rebuild()
SHOPTriggerInputWrite.model_rebuild()
SHOPTriggerInputApply.model_rebuild()
SHOPTriggerOutput.model_rebuild()
SHOPTriggerOutputWrite.model_rebuild()
SHOPTriggerOutputApply.model_rebuild()
Scenario.model_rebuild()
ScenarioWrite.model_rebuild()
ScenarioApply.model_rebuild()
ShopPartialBidCalculationInput.model_rebuild()
ShopPartialBidCalculationInputWrite.model_rebuild()
ShopPartialBidCalculationInputApply.model_rebuild()
ShopPartialBidCalculationOutput.model_rebuild()
ShopPartialBidCalculationOutputWrite.model_rebuild()
ShopPartialBidCalculationOutputApply.model_rebuild()
TaskDispatcherShopInput.model_rebuild()
TaskDispatcherShopInputWrite.model_rebuild()
TaskDispatcherShopInputApply.model_rebuild()
TaskDispatcherShopOutput.model_rebuild()
TaskDispatcherShopOutputWrite.model_rebuild()
TaskDispatcherShopOutputApply.model_rebuild()
TaskDispatcherWaterInput.model_rebuild()
TaskDispatcherWaterInputWrite.model_rebuild()
TaskDispatcherWaterInputApply.model_rebuild()
TaskDispatcherWaterOutput.model_rebuild()
TaskDispatcherWaterOutputWrite.model_rebuild()
TaskDispatcherWaterOutputApply.model_rebuild()
TotalBidMatrixCalculationInput.model_rebuild()
TotalBidMatrixCalculationInputWrite.model_rebuild()
TotalBidMatrixCalculationInputApply.model_rebuild()
TotalBidMatrixCalculationOutput.model_rebuild()
TotalBidMatrixCalculationOutputWrite.model_rebuild()
TotalBidMatrixCalculationOutputApply.model_rebuild()
WaterPartialBidCalculationInput.model_rebuild()
WaterPartialBidCalculationInputWrite.model_rebuild()
WaterPartialBidCalculationInputApply.model_rebuild()
WaterPartialBidCalculationOutput.model_rebuild()
WaterPartialBidCalculationOutputWrite.model_rebuild()
WaterPartialBidCalculationOutputApply.model_rebuild()
Watercourse.model_rebuild()
WatercourseWrite.model_rebuild()
WatercourseApply.model_rebuild()

__all__ = [
    "DataRecord",
    "DataRecordWrite",
    "ResourcesWrite",
    "DomainModel",
    "DomainModelCore",
    "DomainModelWrite",
    "DomainModelList",
    "DomainRelationWrite",
    "ResourcesWriteResult",
    "Alert",
    "AlertWrite",
    "AlertApply",
    "AlertList",
    "AlertWriteList",
    "AlertApplyList",
    "AlertFields",
    "AlertTextFields",
    "BasicBidMatrix",
    "BasicBidMatrixWrite",
    "BasicBidMatrixApply",
    "BasicBidMatrixList",
    "BasicBidMatrixWriteList",
    "BasicBidMatrixApplyList",
    "BasicBidMatrixFields",
    "BasicBidMatrixTextFields",
    "BidCalculationTask",
    "BidCalculationTaskWrite",
    "BidCalculationTaskApply",
    "BidCalculationTaskList",
    "BidCalculationTaskWriteList",
    "BidCalculationTaskApplyList",
    "BidCalculationTaskFields",
    "BidConfiguration",
    "BidConfigurationWrite",
    "BidConfigurationApply",
    "BidConfigurationList",
    "BidConfigurationWriteList",
    "BidConfigurationApplyList",
    "BidConfigurationShop",
    "BidConfigurationShopWrite",
    "BidConfigurationShopApply",
    "BidConfigurationShopList",
    "BidConfigurationShopWriteList",
    "BidConfigurationShopApplyList",
    "BidConfigurationShopFields",
    "BidConfigurationShopTextFields",
    "BidConfigurationWater",
    "BidConfigurationWaterWrite",
    "BidConfigurationWaterApply",
    "BidConfigurationWaterList",
    "BidConfigurationWaterWriteList",
    "BidConfigurationWaterApplyList",
    "BidDocumentAFRR",
    "BidDocumentAFRRWrite",
    "BidDocumentAFRRApply",
    "BidDocumentAFRRList",
    "BidDocumentAFRRWriteList",
    "BidDocumentAFRRApplyList",
    "BidDocumentAFRRFields",
    "BidDocumentAFRRTextFields",
    "BidDocumentDayAhead",
    "BidDocumentDayAheadWrite",
    "BidDocumentDayAheadApply",
    "BidDocumentDayAheadList",
    "BidDocumentDayAheadWriteList",
    "BidDocumentDayAheadApplyList",
    "BidDocumentDayAheadFields",
    "BidDocumentDayAheadTextFields",
    "BidMatrix",
    "BidMatrixWrite",
    "BidMatrixApply",
    "BidMatrixList",
    "BidMatrixWriteList",
    "BidMatrixApplyList",
    "BidMatrixFields",
    "BidMatrixTextFields",
    "BidMatrixRaw",
    "BidMatrixRawWrite",
    "BidMatrixRawApply",
    "BidMatrixRawList",
    "BidMatrixRawWriteList",
    "BidMatrixRawApplyList",
    "BidMatrixRawFields",
    "BidMatrixRawTextFields",
    "BidMethod",
    "BidMethodWrite",
    "BidMethodApply",
    "BidMethodList",
    "BidMethodWriteList",
    "BidMethodApplyList",
    "BidMethodFields",
    "BidMethodTextFields",
    "BidMethodAFRR",
    "BidMethodAFRRWrite",
    "BidMethodAFRRApply",
    "BidMethodAFRRList",
    "BidMethodAFRRWriteList",
    "BidMethodAFRRApplyList",
    "BidMethodAFRRFields",
    "BidMethodAFRRTextFields",
    "BidMethodCustom",
    "BidMethodCustomWrite",
    "BidMethodCustomApply",
    "BidMethodCustomList",
    "BidMethodCustomWriteList",
    "BidMethodCustomApplyList",
    "BidMethodCustomFields",
    "BidMethodCustomTextFields",
    "BidMethodDayAhead",
    "BidMethodDayAheadWrite",
    "BidMethodDayAheadApply",
    "BidMethodDayAheadList",
    "BidMethodDayAheadWriteList",
    "BidMethodDayAheadApplyList",
    "BidMethodDayAheadFields",
    "BidMethodDayAheadTextFields",
    "BidMethodSHOPMultiScenario",
    "BidMethodSHOPMultiScenarioWrite",
    "BidMethodSHOPMultiScenarioApply",
    "BidMethodSHOPMultiScenarioList",
    "BidMethodSHOPMultiScenarioWriteList",
    "BidMethodSHOPMultiScenarioApplyList",
    "BidMethodSHOPMultiScenarioFields",
    "BidMethodSHOPMultiScenarioTextFields",
    "BidMethodWaterValue",
    "BidMethodWaterValueWrite",
    "BidMethodWaterValueApply",
    "BidMethodWaterValueList",
    "BidMethodWaterValueWriteList",
    "BidMethodWaterValueApplyList",
    "BidMethodWaterValueFields",
    "BidMethodWaterValueTextFields",
    "BidRow",
    "BidRowWrite",
    "BidRowApply",
    "BidRowList",
    "BidRowWriteList",
    "BidRowApplyList",
    "BidRowFields",
    "BidRowTextFields",
    "Case",
    "CaseWrite",
    "CaseApply",
    "CaseList",
    "CaseWriteList",
    "CaseApplyList",
    "CaseFields",
    "Commands",
    "CommandsWrite",
    "CommandsApply",
    "CommandsList",
    "CommandsWriteList",
    "CommandsApplyList",
    "CommandsFields",
    "CommandsTextFields",
    "CustomBidMatrix",
    "CustomBidMatrixWrite",
    "CustomBidMatrixApply",
    "CustomBidMatrixList",
    "CustomBidMatrixWriteList",
    "CustomBidMatrixApplyList",
    "CustomBidMatrixFields",
    "CustomBidMatrixTextFields",
    "Generator",
    "GeneratorWrite",
    "GeneratorApply",
    "GeneratorList",
    "GeneratorWriteList",
    "GeneratorApplyList",
    "GeneratorFields",
    "GeneratorTextFields",
    "GeneratorEfficiencyCurve",
    "GeneratorEfficiencyCurveWrite",
    "GeneratorEfficiencyCurveApply",
    "GeneratorEfficiencyCurveList",
    "GeneratorEfficiencyCurveWriteList",
    "GeneratorEfficiencyCurveApplyList",
    "GeneratorEfficiencyCurveFields",
    "Mapping",
    "MappingWrite",
    "MappingApply",
    "MappingList",
    "MappingWriteList",
    "MappingApplyList",
    "MappingFields",
    "MappingTextFields",
    "MarketConfiguration",
    "MarketConfigurationWrite",
    "MarketConfigurationApply",
    "MarketConfigurationList",
    "MarketConfigurationWriteList",
    "MarketConfigurationApplyList",
    "MarketConfigurationFields",
    "MarketConfigurationTextFields",
    "ModelTemplate",
    "ModelTemplateWrite",
    "ModelTemplateApply",
    "ModelTemplateList",
    "ModelTemplateWriteList",
    "ModelTemplateApplyList",
    "ModelTemplateFields",
    "ModelTemplateTextFields",
    "MultiScenarioMatrix",
    "MultiScenarioMatrixWrite",
    "MultiScenarioMatrixApply",
    "MultiScenarioMatrixList",
    "MultiScenarioMatrixWriteList",
    "MultiScenarioMatrixApplyList",
    "MultiScenarioMatrixFields",
    "MultiScenarioMatrixTextFields",
    "MultiScenarioMatrixRaw",
    "MultiScenarioMatrixRawWrite",
    "MultiScenarioMatrixRawApply",
    "MultiScenarioMatrixRawList",
    "MultiScenarioMatrixRawWriteList",
    "MultiScenarioMatrixRawApplyList",
    "MultiScenarioMatrixRawFields",
    "MultiScenarioMatrixRawTextFields",
    "PartialPostProcessingInput",
    "PartialPostProcessingInputWrite",
    "PartialPostProcessingInputApply",
    "PartialPostProcessingInputList",
    "PartialPostProcessingInputWriteList",
    "PartialPostProcessingInputApplyList",
    "PartialPostProcessingInputFields",
    "PartialPostProcessingInputTextFields",
    "PartialPostProcessingOutput",
    "PartialPostProcessingOutputWrite",
    "PartialPostProcessingOutputApply",
    "PartialPostProcessingOutputList",
    "PartialPostProcessingOutputWriteList",
    "PartialPostProcessingOutputApplyList",
    "PartialPostProcessingOutputFields",
    "PartialPostProcessingOutputTextFields",
    "Plant",
    "PlantWrite",
    "PlantApply",
    "PlantList",
    "PlantWriteList",
    "PlantApplyList",
    "PlantFields",
    "PlantTextFields",
    "PlantShop",
    "PlantShopWrite",
    "PlantShopApply",
    "PlantShopList",
    "PlantShopWriteList",
    "PlantShopApplyList",
    "PlantShopFields",
    "PlantShopTextFields",
    "PreprocessorInput",
    "PreprocessorInputWrite",
    "PreprocessorInputApply",
    "PreprocessorInputList",
    "PreprocessorInputWriteList",
    "PreprocessorInputApplyList",
    "PreprocessorInputFields",
    "PreprocessorInputTextFields",
    "PreprocessorOutput",
    "PreprocessorOutputWrite",
    "PreprocessorOutputApply",
    "PreprocessorOutputList",
    "PreprocessorOutputWriteList",
    "PreprocessorOutputApplyList",
    "PreprocessorOutputFields",
    "PreprocessorOutputTextFields",
    "PriceArea",
    "PriceAreaWrite",
    "PriceAreaApply",
    "PriceAreaList",
    "PriceAreaWriteList",
    "PriceAreaApplyList",
    "PriceAreaFields",
    "PriceAreaTextFields",
    "PriceAreaAFRR",
    "PriceAreaAFRRWrite",
    "PriceAreaAFRRApply",
    "PriceAreaAFRRList",
    "PriceAreaAFRRWriteList",
    "PriceAreaAFRRApplyList",
    "PriceAreaAFRRFields",
    "PriceAreaAFRRTextFields",
    "PriceAreaAsset",
    "PriceAreaAssetWrite",
    "PriceAreaAssetApply",
    "PriceAreaAssetList",
    "PriceAreaAssetWriteList",
    "PriceAreaAssetApplyList",
    "PriceAreaAssetFields",
    "PriceAreaAssetTextFields",
    "PriceProdCase",
    "PriceProdCaseWrite",
    "PriceProdCaseApply",
    "PriceProdCaseList",
    "PriceProdCaseWriteList",
    "PriceProdCaseApplyList",
    "PriceProdCaseFields",
    "Reservoir",
    "ReservoirWrite",
    "ReservoirApply",
    "ReservoirList",
    "ReservoirWriteList",
    "ReservoirApplyList",
    "ReservoirFields",
    "ReservoirTextFields",
    "SHOPResult",
    "SHOPResultWrite",
    "SHOPResultApply",
    "SHOPResultList",
    "SHOPResultWriteList",
    "SHOPResultApplyList",
    "SHOPResultFields",
    "SHOPResultPriceProd",
    "SHOPResultPriceProdWrite",
    "SHOPResultPriceProdApply",
    "SHOPResultPriceProdList",
    "SHOPResultPriceProdWriteList",
    "SHOPResultPriceProdApplyList",
    "SHOPResultPriceProdFields",
    "SHOPTimeSeries",
    "SHOPTimeSeriesWrite",
    "SHOPTimeSeriesApply",
    "SHOPTimeSeriesList",
    "SHOPTimeSeriesWriteList",
    "SHOPTimeSeriesApplyList",
    "SHOPTimeSeriesFields",
    "SHOPTimeSeriesTextFields",
    "SHOPTriggerInput",
    "SHOPTriggerInputWrite",
    "SHOPTriggerInputApply",
    "SHOPTriggerInputList",
    "SHOPTriggerInputWriteList",
    "SHOPTriggerInputApplyList",
    "SHOPTriggerInputFields",
    "SHOPTriggerInputTextFields",
    "SHOPTriggerOutput",
    "SHOPTriggerOutputWrite",
    "SHOPTriggerOutputApply",
    "SHOPTriggerOutputList",
    "SHOPTriggerOutputWriteList",
    "SHOPTriggerOutputApplyList",
    "SHOPTriggerOutputFields",
    "SHOPTriggerOutputTextFields",
    "Scenario",
    "ScenarioWrite",
    "ScenarioApply",
    "ScenarioList",
    "ScenarioWriteList",
    "ScenarioApplyList",
    "ScenarioFields",
    "ScenarioTextFields",
    "ShopPartialBidCalculationInput",
    "ShopPartialBidCalculationInputWrite",
    "ShopPartialBidCalculationInputApply",
    "ShopPartialBidCalculationInputList",
    "ShopPartialBidCalculationInputWriteList",
    "ShopPartialBidCalculationInputApplyList",
    "ShopPartialBidCalculationInputFields",
    "ShopPartialBidCalculationInputTextFields",
    "ShopPartialBidCalculationOutput",
    "ShopPartialBidCalculationOutputWrite",
    "ShopPartialBidCalculationOutputApply",
    "ShopPartialBidCalculationOutputList",
    "ShopPartialBidCalculationOutputWriteList",
    "ShopPartialBidCalculationOutputApplyList",
    "ShopPartialBidCalculationOutputFields",
    "ShopPartialBidCalculationOutputTextFields",
    "TaskDispatcherShopInput",
    "TaskDispatcherShopInputWrite",
    "TaskDispatcherShopInputApply",
    "TaskDispatcherShopInputList",
    "TaskDispatcherShopInputWriteList",
    "TaskDispatcherShopInputApplyList",
    "TaskDispatcherShopInputFields",
    "TaskDispatcherShopInputTextFields",
    "TaskDispatcherShopOutput",
    "TaskDispatcherShopOutputWrite",
    "TaskDispatcherShopOutputApply",
    "TaskDispatcherShopOutputList",
    "TaskDispatcherShopOutputWriteList",
    "TaskDispatcherShopOutputApplyList",
    "TaskDispatcherShopOutputFields",
    "TaskDispatcherShopOutputTextFields",
    "TaskDispatcherWaterInput",
    "TaskDispatcherWaterInputWrite",
    "TaskDispatcherWaterInputApply",
    "TaskDispatcherWaterInputList",
    "TaskDispatcherWaterInputWriteList",
    "TaskDispatcherWaterInputApplyList",
    "TaskDispatcherWaterInputFields",
    "TaskDispatcherWaterInputTextFields",
    "TaskDispatcherWaterOutput",
    "TaskDispatcherWaterOutputWrite",
    "TaskDispatcherWaterOutputApply",
    "TaskDispatcherWaterOutputList",
    "TaskDispatcherWaterOutputWriteList",
    "TaskDispatcherWaterOutputApplyList",
    "TaskDispatcherWaterOutputFields",
    "TaskDispatcherWaterOutputTextFields",
    "TotalBidMatrixCalculationInput",
    "TotalBidMatrixCalculationInputWrite",
    "TotalBidMatrixCalculationInputApply",
    "TotalBidMatrixCalculationInputList",
    "TotalBidMatrixCalculationInputWriteList",
    "TotalBidMatrixCalculationInputApplyList",
    "TotalBidMatrixCalculationInputFields",
    "TotalBidMatrixCalculationInputTextFields",
    "TotalBidMatrixCalculationOutput",
    "TotalBidMatrixCalculationOutputWrite",
    "TotalBidMatrixCalculationOutputApply",
    "TotalBidMatrixCalculationOutputList",
    "TotalBidMatrixCalculationOutputWriteList",
    "TotalBidMatrixCalculationOutputApplyList",
    "TotalBidMatrixCalculationOutputFields",
    "TotalBidMatrixCalculationOutputTextFields",
    "TurbineEfficiencyCurve",
    "TurbineEfficiencyCurveWrite",
    "TurbineEfficiencyCurveApply",
    "TurbineEfficiencyCurveList",
    "TurbineEfficiencyCurveWriteList",
    "TurbineEfficiencyCurveApplyList",
    "TurbineEfficiencyCurveFields",
    "WaterPartialBidCalculationInput",
    "WaterPartialBidCalculationInputWrite",
    "WaterPartialBidCalculationInputApply",
    "WaterPartialBidCalculationInputList",
    "WaterPartialBidCalculationInputWriteList",
    "WaterPartialBidCalculationInputApplyList",
    "WaterPartialBidCalculationInputFields",
    "WaterPartialBidCalculationInputTextFields",
    "WaterPartialBidCalculationOutput",
    "WaterPartialBidCalculationOutputWrite",
    "WaterPartialBidCalculationOutputApply",
    "WaterPartialBidCalculationOutputList",
    "WaterPartialBidCalculationOutputWriteList",
    "WaterPartialBidCalculationOutputApplyList",
    "WaterPartialBidCalculationOutputFields",
    "WaterPartialBidCalculationOutputTextFields",
    "Watercourse",
    "WatercourseWrite",
    "WatercourseApply",
    "WatercourseList",
    "WatercourseWriteList",
    "WatercourseApplyList",
    "WatercourseFields",
    "WatercourseTextFields",
    "WatercourseShop",
    "WatercourseShopWrite",
    "WatercourseShopApply",
    "WatercourseShopList",
    "WatercourseShopWriteList",
    "WatercourseShopApplyList",
    "WatercourseShopFields",
    "WatercourseShopTextFields",
]
